# Copyright (c) 2020, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Callable


def callback(callback: Optional[Callable[[int], None]], /) -> None:
    """
    Register a callback method that is called whenever a modem status is
    generated by the XBee.

    This callback takes one parameter, the integer value of the modem status.

    **Note: This is only available on XBee 3 RF products with version
    ending in 0B or newer, and XBee 3 Cellular products with version
    ending in 1B or newer.**

    :param callback: A function that is called whenever a modem status is
        received.
        If ``callback`` is None, the registered callback will be
        unregistered, allowing for polling of modem_status.receive() again.
    """
    ...


def receive() -> Optional[int]:
    """
    Returns a modem status if one is available. This function can only
    be called if no callback is registered, otherwise it will raise an
    exception.

    **Note: This is only available on XBee 3 RF products with version
    ending in 0B or newer, and XBee Cellular and XBee 3 Cellular products
    with version ending in 1B or newer.**

    :return: The integer value of a modem status, if available.
    """
    ...



import utime
import ustruct
import machine 

from micropython import const


class DIO:
    """
    The DIO class provides static methods required to 
    define DIO port pins for the AutoCtrl.
    """
    
    IN = machine.Pin.IN
    OUT = machine.Pin.OUT
    PULL_UP = machine.Pin.PULL_UP
    PULL_DOWN = machine.Pin.PULL_DOWN
    LOW = const(0)
    HIGH = const(1)

    class Device:
        """
        The Device class is used to specify the device type.
        """
        RELAY = const(0) 
        PWM = const(1)    
    
    P_RELAY = {0:'D0', 1:'D6', 2:'D5'} 
    
    @staticmethod
    def P18() -> machine.Pin:
        """
        Create a pin object for the P18 port of DIO.
        
        Only Input of ActiveHigh(5V ~ 6V) Device(ex: GasDetector). 
        It has a built-in divider resistor that halves the 12V.
        
        :return: The pin object.
        """
    
    @staticmethod
    def P17() -> machine.Pin:
        """
        Create a pin object for the P17 port of DIO. 
        
        Only Input of ActiveLow(GND) Device(ex:PIR, LimitSiwtch ...)
        
        :return: The pin object.
        """
     
    @staticmethod
    def P8(mode:int, pull_value:int=None) -> machine.Pin:
        """
        Create a pin object for the P8 port of DIO.
        
        Operation 3V3 IN/OUT
        
        :param mode: The mode of the pin. DIO.IN or DIO.OUT
        :param pull_value: The pull value of the pin. DIO.PULL_UP or DIO.PULL_DOWN
        
        :return: The pin object.
        """
    
    @staticmethod
    def P23(mode, pull_value=None):
        """
        Create a pin object for the P23 port of DIO.
        
        Operation 3V3 IN/OUT
        
        :param mode: The mode of the pin. DIO.IN or DIO.OUT
        :param pull_value: The pull value of the pin. DIO.PULL_UP or DIO.PULL_DOWN
        
        :return: The pin object.
        """

def Relay(pin, value=0):
    """
    Create a relay object.
    
    :param pin: The pin string of the relay. DIO.P_RELAY[0] ~ DIO.P_RELAY[2]
    :param value: The initial value of the relay. value=0 is Normal open
    
    :return: The relay object.
    """    

class RelayTerminal:
    """
    The RelayTerminal object is used to control the state of the RelayTerminal.
    """
    
    def __init__(self, *relays) -> None:
        """
        Initializes the RelayTerminal object.
        
        :param relays: The relay objects to control the RelayTerminal.
        """
        
    def on(self, pos:int) -> None:
        """
        Turns on the relay.
        
        :param ch: The index number of the relay to turn on.
        """
        
    def off(self, pos:int) -> None:
        """
        Turns off the relay.
        
        :param pos: The index number of the relay to turn off.
        """
        
class DoorLock:
    """
    The DoorLock object is used to control the state of the DoorLock.
    """
    
    def __init__(self, relay, *, dio:int, active_low:bool) -> None:
        """
        Initializes the DoorLock object.
        dio is only Input (ActiveHigh). DIO Pin object (ex P17)
        
        :param relay: The relay object to control the DoorLock.
        :param dio: The feedback pin to read the state of the DoorLock.
        :param active_low: The state of the feedback pin when the DoorLock is opened.        
        
        The open, close, and is_opened methods can only be used when the dio and active_low parameters are given.
        """

    def open(self) -> None:
        """
        Opens the DoorLock when dio and active_low parameters are given.
        """
    
    def close(self) -> None:
        """
        Closes the DoorLock when dio and active_low parameters are given.
        """

    def is_opened(self) -> bool:
        """
        Returns the state of the DoorLock when dio and active_low parameters are given.
        
        :return: ``True`` if the DoorLock is opened, ``False`` if the DoorLock is closed.   
        """
    
    def work(self) -> None:
        """
        Changes the status of the DoorLock.
        """

class Pwm:
    """
    The Pwm object is used to control the state of the PWM.
    """
    
    def __init__(self, freq:int=100) -> None:
        """
        Initializes the Pwm object.
        
        :param freq: The frequency of the PWM signal.
        """
     
    def freq(self, freq: int) -> None:
        """
        Set the frequency of the PWM signal.
        
        :param freq: The frequency of the PWM signal. Maximum frequency is 1526 Hz.
        """

    def pwm(self, ch: int, on: int, off: int) -> None:
        """
        Set the PWM signal.
        
        :param ch: The channel number of the PWM signal.
        :param on: The time the signal is on.
        :param off: The time the signal is off.
        
        duty cycle 0: on is 0, off is 4096
        duty cycle 100: on is 4096, off is 0
        etc: on is 0, off is int((duty_cycle - 1) * (4095 - 1) / (99 - 1) + 1)
        """

    def duty(self, ch:int, value:int) -> None:
        """
        Set the duty cycle of the PWM signal.
        
        :param ch: The channel number of the PWM signal.
        :param value: The duty cycle of the PWM signal.
        """

class GasDetector:
    """
    The GasDetector object is used to read the state of the GasDetector.
    """
    
    def read(self):
        """
        read the state of the GasDetector.
        
        :return: The state of the GasDetector.
        """

class GasBreaker: 
    """
    The GasBreaker object is used to control the state of the GasBreaker.
    """    
    
    def __init__(self, red, black) -> None:
        """
        Initializes the GasBreaker object.
        
        :param red: Channel number to which the red wire of the gas breaker is connected.
        :param black: Channel number to which the black wire of the gas breaker is connected.
        """        
        
    def open(self) -> None:
        """
        Opens the gas breaker.
        """
            
    def close(self, init=False):
        """
        Closes the gas breaker.
        """
    
    def stop(self) -> None:
        """
        Stops the gas breaker.
        """
    
def Fan(type, ch, value:int=0) -> object:
    """
    The Fan object is used to control the state of the Fan.
    
    :param type: The device to control, either DIO.Device.RELAY or DIO.Device.PWM
    :param ch: The channel number of the device to control
    :param value: Initial value
    
    :return: Relay or PWM object
    """
     
def Light(type, ch, value=0) -> object:
    """
    The Light object is used to control the state of the Light.
    
    :param type: The device to control, either DIO.Device.RELAY or DIO.Device.PWM
    :param ch: The channel number of the device to control
    :param value: Initial value
    
    :return: Relay or PWM object
    """



import os
import sys
import glob
import time
import re
import platform
import threading
import posixpath
import ast
import textwrap
import binascii
import shutil

import click
import dotenv
import serial

from genlib.ansiec import ANSIEC
 

from . import __version__  

#--------------------------------------------------------------

def windows_full_port_name(portname):
    m = re.match(r"^COM(\d+)$", portname)
    if m and int(m.group(1)) < 10:
        return portname
    else:
        return "\\\\.\\{0}".format(portname)

def is_micropython_board(port):
    port = port.upper()
    
    if platform.system() == "Windows":
        port = windows_full_port_name(port)
        
    try:
        with serial.Serial(port, 115200, timeout=1) as ser:
            ser.write(b'\x03') # Ctrl + C(b'\x03') --> interrupt any running program
            ser.read_all().decode('utf-8') # >>>
            ser.write(b'\x02') # Ctrl + B(b'\x02') --> enter normal repl, ref: Ctrl + D(b'\x04') --> soft reset, 
            time.sleep(0.1)
            response = ser.read_all().decode('utf-8').strip()

            if 'MicroPython' in response:
                s = response.find("MicroPython") + len("MicroPython")
                e = response.find('Type "help()"')
                return response[s:e].strip()
    except (OSError, serial.SerialException):
        pass
    
    return None

def device_scan():
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]    
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    color_tbl = (ANSIEC.FG.BRIGHT_YELLOW, ANSIEC.FG.BRIGHT_GREEN, ANSIEC.FG.BRIGHT_BLUE)
    color_pos = 0    
    
    for port in ports:
        descript = is_micropython_board(port)
        if descript:
            print(color_tbl[color_pos] + f"{port}" + ANSIEC.OP.RESET + f" ({descript})")
            color_pos = (color_pos + 1) % len(color_tbl)

def host_environment(board):
    import upyboard
    if board in ('xnode', 'pico2'):
        _board_lib_path = os.path.join(os.path.dirname(upyboard.__file__), board) 
    else:    
        print("The device type " + ANSIEC.FG.BRIGHT_RED + f"{board}" + ANSIEC.OP.RESET + " is not supported.")
        return
    
    vscode_dir = ".vscode" 
    typehints_src = os.path.join(_board_lib_path, "typehints")

    upyboard_file = os.path.join(vscode_dir, ".upyboard")
    task_file = os.path.join(vscode_dir, "tasks.json") 
    settings_file = os.path.join(vscode_dir, "settings.json") 
    
    task_file_contents = """{
    "version": "2.0.0",
    "tasks": [
        {
            "label": "Run micropython with upyboard",
            "type": "shell",
            "command": "upy",
            "args": [
                "${file}"
            ],
            "problemMatcher": [],
            "group": {
                "kind": "build",
                "isDefault": true
            }
        }
    ]
}
"""

    settings_file_contents = """{
    "python.languageServer": "Pylance",
    "python.analysis.diagnosticSeverityOverrides": {
        "reportMissingModuleSource": "none",
    },
    "python.analysis.extraPaths": [
        "./.vscode"
    ]
}
"""

    if os.path.exists(vscode_dir):
        print(f"The {vscode_dir} folder already exists. This folder will be deleted to proceed. Do you wish to continue? (y or n): ", end='')
        try:
            if input().strip().lower() == 'n':
                print("Canceling the operation")
                return
        except KeyboardInterrupt:
            print("Canceling the operation")
            return
        
        shutil.rmtree(vscode_dir, ignore_errors=True)
        
    shutil.copytree(typehints_src, vscode_dir) 
        
    with open(upyboard_file, "w") as f:
        f.write("SERIAL_PORT=\n")
        f.write("DEVICE_TYPE=xbee3\n")

    with open(task_file, "w") as f:
        f.write(task_file_contents)  

    with open(settings_file, "w") as f:
        f.write(settings_file_contents)  

    print(f"Please open the {ANSIEC.FG.BRIGHT_GREEN}.upyboard file{ANSIEC.OP.RESET} within the {ANSIEC.FG.BRIGHT_BLUE}.vscode folder{ANSIEC.OP.RESET} and configure the {ANSIEC.FG.BRIGHT_YELLOW}SERIAL_PORT{ANSIEC.OP.RESET} setting.")


buffer = b''
expected_bytes = 0

def stdout_write_bytes(b):
    global buffer, expected_bytes
    
    if b == b'\x04':
        return
    
    if expected_bytes > 0: 
        buffer += b
        expected_bytes -= 1

        if expected_bytes == 0: 
            try:
                sys.stdout.buffer.write(buffer)
                sys.stdout.buffer.flush()
            except UnicodeDecodeError:
                sys.stdout.buffer.write(buffer.hex())
            finally:
                buffer = b'' 
    elif ord(b) <= 0x7F:                # ASCII
        sys.stdout.buffer.write(b)
        sys.stdout.buffer.flush()
    else:                               # Multi-byte
        if (ord(b) & 0xF0) == 0xF0:     # 4 byte
            expected_bytes = 3
        elif (ord(b) & 0xE0) == 0xE0:   # 3 byte
            expected_bytes = 2
        elif (ord(b) & 0xC0) == 0xC0:   # 2 byte
            expected_bytes = 1
        else:
            sys.stdout.buffer.write(buffer.hex())
            return 

        buffer = b                      # save first byte  
    
#--------------------------------------------------------------

class Board:
    BUFFER_SIZE = 32

    def __init__(self, port, baudrate=115200, wait=0):     
        self.in_raw_repl = False
        self.use_raw_paste = True
        
        if platform.system() == "Windows":
            port = windows_full_port_name(port)
            
        delayed = False
        for attempt in range(wait + 1):
            try:
                self.serial = serial.Serial(port, baudrate, inter_byte_timeout=0.1, exclusive=True)
                self.serial.dtr = False
                self.serial.rts = False
                break
            except (OSError, IOError): 
                if wait == 0:
                    continue
                if attempt == 0:
                    sys.stdout.write(f"Waiting {wait} seconds for board ")
                    delayed = True
            time.sleep(1)
            sys.stdout.write('.')
            sys.stdout.flush()
        else:
            if delayed:
                print('')
            raise BaseException(f"failed to access {port}")
        if delayed:
            print('')
        
        self.__init_repl()

    def __init_repl(self):
        self.serial_reader_running = None
        self.serial_out_put_enable = True
        self.serial_out_put_count = 0

    def close(self):
        self.serial.close()

    def read_until(self, min_num_bytes, ending, timeout=10, data_consumer=None):
        assert data_consumer is None or len(ending) == 1
        
        data = self.serial.read(min_num_bytes)
        
        if data_consumer:
            data_consumer(data)
        timeout_count = 0
        
        while True:
            if data.endswith(ending):
                break
            elif self.serial.in_waiting > 0:
                new_data = self.serial.read(1)
                if data_consumer:
                    data_consumer(new_data)
                    data = new_data
                else:                
                    data = data + new_data
                timeout_count = 0
            else:
                timeout_count += 1
                if timeout is not None and timeout_count >= 100 * timeout:
                    break
                time.sleep(0.01)
        return data

    def enter_raw_repl(self, soft_reset=True):
        self.serial.write(b'\r\x03') # ctrl-C: interrupt any running program

        n = self.serial.in_waiting
        while n > 0:
            self.serial.read(n)
            n = self.serial.in_waiting

        self.serial.write(b'\r\x01') # ctrl-A: enter raw REPL
        
        if soft_reset:
            data = self.read_until(1, b'raw REPL; CTRL-B to exit\r\n>')
            if not data.endswith(b'raw REPL; CTRL-B to exit\r\n>'):
                print(data)
                raise BaseException('could not enter raw repl')

            self.serial.write(b'\x04') # ctrl-D: soft reset
            
            data = self.read_until(1, b'soft reboot\r\n')
            if not data.endswith(b'soft reboot\r\n'):
                print(data)
                raise BaseException('could not enter raw repl')

        data = self.read_until(1, b'raw REPL; CTRL-B to exit\r\n')
        if not data.endswith(b'raw REPL; CTRL-B to exit\r\n'):
            print(data)
            raise BaseException('could not enter raw repl')
        
        self.in_raw_repl = True
        
    def exit_raw_repl(self):
        self.serial.write(b'\r\x02') # ctrl-B: enter friendly REPL
        self.in_raw_repl = False
        
    def _follow_write(self, echo):        
        try:
            import msvcrt
            def getkey():
                return msvcrt.getch()

            def putkey(ch):
                if ch == b'\r':
                    ch = b'\n'
                msvcrt.putch(ch)
                
        except ImportError:
            import sys, tty, termios
            def getkey():
                fd = sys.stdin.fileno()
                old = termios.tcgetattr(fd)
                try:
                    tty.setraw(sys.stdin.fileno())
                    ch = sys.stdin.read(1)
                finally:
                    termios.tcsetattr(fd, termios.TCSADRAIN, old)
                return ch
            
            def putkey(ch):
                sys.stdout.write(ch)
                sys.stdout.flush()
        
        while True:
            ch = getkey()
            if ch == b'\x03': # Ctrl + C
                os._exit(0)
            if echo:
                putkey(ch)
            self.serial.write(ch)

    def follow(self, timeout, data_consumer=None, input_stat=None):
        if input_stat[1]:
            threading.Thread(target=self._follow_write, args=(input_stat[0],), daemon=True).start()
        
        data = self.read_until(1, b'\x04', timeout=timeout, data_consumer=data_consumer)
        if not data.endswith(b'\x04'):
            raise BaseException('timeout waiting for first EOF reception')
        data = data[:-1]
        
        data_err = self.read_until(1, b'\x04', timeout=timeout)
        if not data_err.endswith(b'\x04'):
            raise BaseException('timeout waiting for second EOF reception')
        data_err = data_err[:-1]
        
        return data, data_err

    def exec_raw_no_follow(self, command):            
        if isinstance(command, bytes):
            command_bytes = command
        else:
            command_bytes = bytes(command, encoding='utf8')

        data = self.read_until(1, b'>')
        if not data.endswith(b'>'):
            raise BaseException('could not enter raw repl')

        for i in range(0, len(command_bytes), 256):
            self.serial.write(command_bytes[i:min(i + 256, len(command_bytes))])
            time.sleep(0.01)
        self.serial.write(b'\x04')

        data = self.read_until(1, b'OK')
        if not data.endswith(b'OK'):
            raise BaseException('could not exec commandm (response: %r)' % data)

    def exec_raw(self, command, timeout=None, data_consumer=None, input_stat=None):
        self.exec_raw_no_follow(command)
        return self.follow(timeout, data_consumer, input_stat)

    def exec(self, command, stream_output=False, echo_on=False):
        data_consumer = None
        if stream_output or echo_on:
            data_consumer = stdout_write_bytes
        ret, ret_err = self.exec_raw(command, data_consumer=data_consumer, input_stat=(stream_output, echo_on))
        if ret_err:
            raise BaseException(ret_err.decode('utf-8'))
        return ret
    
    def execfile(self, filename, stream_output=False, echo_on=False):
        with open(filename, 'r+b') as f:
            command = f.read()
        return self.exec(command, stream_output, echo_on)
         
    def _exec_command(self, command):
        self.enter_raw_repl()
        try:
            out = self.exec(textwrap.dedent(command))
        except BaseException as ex:
            raise ex
        self.exit_raw_repl()
        return out

    def run(self, filename, stream_output=False, echo_on=False):
        self.enter_raw_repl()

        if not stream_output and not echo_on:           # -n
            with open(filename, "rb") as infile:            # Running without io stream
                self.exec_raw_no_follow(infile.read())
        elif not stream_output and echo_on:             # -in (default)
            self.execfile(filename, False, True)            
        elif stream_output and echo_on:                 # -i (echo on)
            self.execfile(filename, True, True)                         
        else:                                           # default (-in)
            self.execfile(filename, False, True)            

        self.exit_raw_repl()

    def __repl_serial_to_stdout(self):        
        def hexsend(string_data=''):
            import binascii
            hex_data = binascii.unhexlify(string_data)
            return hex_data

        try:
            data = b''
            try:
                while self.serial_reader_running:
                    count = self.serial.in_waiting
                    if count == 0:
                        time.sleep(0.01)
                        continue

                    if count > 0:
                        data += self.serial.read(count)

                        if len(data) < 20:
                            try:
                                data.decode()
                            except UnicodeDecodeError:
                                continue

                        if data != b'':
                            if self.serial_out_put_enable and self.serial_out_put_count > 0:
                                if platform.system() == 'Windows':   
                                    sys.stdout.buffer.write(data.replace(b"\r", b""))
                                else:
                                    sys.stdout.buffer.write(data)
                                    
                                sys.stdout.buffer.flush()
                        else:
                            self.serial.write(hexsend(data))

                        data = b''
                        self.serial_out_put_count += 1
            except:
                print('')
                return
        except KeyboardInterrupt:
            if serial != None:
                serial.close()
    
    def reset(self):
        command = f"""
            import machine
            machine.soft_reset()
        """
        self._exec_command(command)
    
    def repl(self):
        self.serial_reader_running = True
        self.serial_out_put_enable = True
        self.serial_out_put_count = 1

        self.reset()
        self.read_until(1, b'\x3E\x3E\x3E', timeout=1) # read prompt >>>

        repl_thread = threading.Thread(target=self.__repl_serial_to_stdout, daemon=True, name='REPL')
        repl_thread.start()

        if platform.system() == 'Windows':   
            import msvcrt as getch
        else:
            import getch
            
        self.serial.write(b'\r') # Update prompt
        
        count = 0
        
        while True:
            char = getch.getch()
        
            if char == b'\x16': # Ctrl + V(\x16) to Ctrl + C(\x03)
                char = b'\x03'

            count += 1
            if count == 1000:
                time.sleep(0.01)
                count = 0

            if char == b'\x07':
                self.serial_out_put_enable = False
                continue

            if char == b'\x0F':
                self.serial_out_put_enable = True
                self.serial_out_put_count = 0
                continue

            if char == b'\x00' or not char:
                continue

            if char == b'\x18':   # Ctrl + X to exit repl mode
                self.serial_reader_running = False
                self.serial.write(b' ')
                time.sleep(0.01)
                print('')
                break
            try:
                self.serial.write(b'\r' if char == b'\n' else char)
            except:
                print('')
                break
           
    def fs_get(self, filename):
        command = f"""
            import sys
            import ubinascii
            with open('{filename}', 'rb') as infile:
                while True:
                    result = infile.read({self.BUFFER_SIZE})
                    if result == b'':
                        break
                    len = sys.stdout.write(ubinascii.hexlify(result))
        """
        out = self._exec_command(command)
        return binascii.unhexlify(out)

    def fs_ls(self, dir="/"):
        if not dir.startswith("/"):
            dir = "/" + dir
        #if dir.endswith("/"):
        #    dir = dir[:-1]
            
        command = f"""
            import os
            def listdir(dir):
                if dir == '/':                
                    return sorted([dir + f for f in os.listdir(dir)])
                else:
                    return sorted([dir + '/' + f for f in os.listdir(dir)])
            print(listdir('{dir}'))
        """
        out = self._exec_command(command)
        return ast.literal_eval(out.decode("utf-8"))
            
    def fs_is_dir(self, path):
        command = f"""
            vstat = None
            try:
                from os import stat
            except ImportError:
                from os import listdir
                vstat = listdir
            def ls_dir(path):
                if vstat is None:
                    return stat(path)[0] & 0x4000 != 0
                else:
                    try:
                        vstat(path)
                        return True
                    except OSError as e:
                        return False
            print(ls_dir('{path}'))
        """
        out = self._exec_command(command)
        return ast.literal_eval(out.decode("utf-8"))

    def fs_mkdir(self, dir):       
        command = f"""
            import os
            def mkdir(dir):
                parts = dir.split(os.sep)
                dirs = [os.sep.join(parts[:i+1]) for i in range(len(parts))]
                check = 0
                for d in dirs:
                    try:
                        os.mkdir(d)
                    except OSError as e:
                        check += 1
                        if "EEXIST" in str(e):
                            continue
                        else:
                            return False
                return check < len(parts)
            print(mkdir('{dir}'))
        """        
        out = self._exec_command(command)
        return ast.literal_eval(out.decode("utf-8"))

    def fs_putdir(self, local, remote, callback=None):        
        for parent, child_dirs, child_files in os.walk(local, followlinks=True):
            remote_parent = posixpath.normpath(posixpath.join(remote, os.path.relpath(parent, local)))
           
            try:
                self.fs_mkdir(remote_parent)
            except:
                pass
        
            for filename in child_files:
                with open(os.path.join(parent, filename), "rb") as infile:
                    remote_filename = posixpath.join(remote_parent, filename)
                    data = infile.read()

                    total_size = os.path.getsize(os.path.join(parent, filename))                 
                    if callback:
                        th = threading.Thread(target=callback, args=(remote_filename, total_size), daemon=True)
                        th.start()
                        
                    self.fs_put(data, remote_filename)
                    
                    if callback:
                        th.join() 

    def fs_put(self, local_data, remote, callback=None):
        self.enter_raw_repl()
        try:
            self.exec(f"f = open('{remote}', 'wb')")
        except BaseException as e:
            if "EEXIST" in str(e):
                self.exit_raw_repl()
                self.fs_rm(remote)
                self.fs_put(local_data, remote, callback)
            return

        size = len(local_data)
        if callback:
            th = threading.Thread(target=callback, args=(remote, size), daemon=True)
            th.start()
            
        for i in range(0, size, self.BUFFER_SIZE):
            chunk_size = min(self.BUFFER_SIZE, size - i)
            chunk = repr(local_data[i : i + chunk_size])
            if not chunk.startswith("b"):
                chunk = "b" + chunk
            self.exec(f"f.write({chunk})")
        
        self.exec("f.close()")
        self.exit_raw_repl()
        
        if callback:
            th.join() 

    def fs_rm(self, filename):
        command = f"""
            import os
            os.remove('{filename}')
        """
        self._exec_command(command)

    def fs_rmdir(self, dir):
        command = f"""
            import os
            def rmdir(dir):
                os.chdir(dir)
                for f in os.listdir():
                    try:
                        os.remove(f)
                    except OSError:
                        pass
                for f in os.listdir():
                    rmdir(f)
                os.chdir('..')
                os.rmdir(dir)
            rmdir('{dir}')
        """
        self._exec_command(command)

    def fs_format(self, type):
        ret = True
        
        if type == "lopy":
            command = """ 
                import os
                os.fsformat('/flash')
            """
        elif type == "xbee3":
            command = """
                import os
                os.format()
            """
        elif type == "pico2":
            command = """
                import os
                import rp2
                bdev = rp2.Flash()
                os.VfsFat.mkfs(bdev)
            """
        else:
            ret = False
        try:
            self._exec_command(command)
        except BaseException:
            ret = False
            
        return ret


#--------------------------------------------------------------

config = dotenv.find_dotenv(filename=".vscode/.upyboard", usecwd=True)
if config:
    dotenv.load_dotenv(dotenv_path=config)

_board = None
_type = None
_root_fs = None
_board_lib_path = None
_sport = None

@click.group()
@click.option(
    "--sport",
    "-s",
    envvar="SERIAL_PORT",
    required=True,
    type=click.STRING,
    help="Serial port name for connected board.",
    metavar="SPORT",
)
@click.option(
    "--baud",
    '-b',
    envvar="SERIAL_BAUD",
    default=115200,
    type=click.INT,
    help="Baud rate for the serial connection (default 115200).",
    metavar="BAUD",
)
@click.option(
    "--type",
    '-t',
    envvar="DEVICE_TYPE",
    default='xbee3',
    type=click.STRING,
    help="Device type",
    metavar="TYPE",
)
def upyboard(sport, baud, type):
    global _board, _type, _root_fs, _board_lib_path, _sport
    
    try:
        _sport = sport
        _board = Board(_sport, baud)
    except BaseException as ex:
        print("Board is not connected to " + ANSIEC.FG.BRIGHT_RED + f"{sport}" + ANSIEC.OP.RESET)
        print("Please check the ports with the scan command and try again.")
        raise click.Abort()

    _type = type.lower().strip()
    
    if _type == 'xbee3':
        _root_fs = "/flash/"
        _board_lib_path = os.path.join(os.path.dirname(__file__), "xnode") 
    elif _type == 'pico2':
        _root_fs = "/"
        _board_lib_path = os.path.join(os.path.dirname(__file__), _type) 
    else:    
        print("The device type " + ANSIEC.FG.BRIGHT_RED + f"{_type}" + ANSIEC.OP.RESET + " is not supported.")
        raise click.Abort()
        
@upyboard.command()
@click.argument("remote_file")
@click.argument("local_file", type=click.File("wb"), required=False)
def get(remote_file, local_file):
    if not remote_file.startswith(_root_fs):
        remote_file = _root_fs + remote_file
    
    try:
        contents = _board.fs_get(remote_file)
    
        if local_file is None:
            try:
                print(contents.decode("utf-8"))
            except:
                print(f"{contents}")
        else:
            local_file.write(contents)
    except BaseException:
        remote_file = remote_file.replace(_root_fs, "", 1)
        print("The file " + ANSIEC.FG.BRIGHT_RED + f"{remote_file}" + ANSIEC.OP.RESET + " does not exist.")
    
@upyboard.command()
@click.argument("dir")
def mkdir(dir):
    dir_ = dir
    if not dir_.startswith(_root_fs):
        dir_ = _root_fs + dir
        
    if _board.fs_mkdir(dir_):
        print(f"{dir} is " + ANSIEC.FG.BRIGHT_GREEN + "created." + ANSIEC.OP.RESET)
    else:
        print(f"{dir} is " + ANSIEC.FG.BRIGHT_RED + "already exists." + ANSIEC.OP.RESET)

@upyboard.command()
@click.argument("remote")
def rm(remote):
    if not remote.startswith(_root_fs):
        remote = _root_fs + remote
        
    try:
        if _board.fs_is_dir(remote):
            _board.fs_rmdir(remote)
        else:
            _board.fs_rm(remote)
    except BaseException:
        remote = remote.replace(_root_fs, "", 1)
        print("The " + ANSIEC.FG.BRIGHT_RED + f"{remote}" + ANSIEC.OP.RESET + " does not exist.")
        
@upyboard.command()
@click.argument("dir", default="/")
def ls(dir):          
    if not dir.startswith(_root_fs):
        dir = _root_fs + dir
            
    try:
        for f in _board.fs_ls(dir):
            f_name = f.split("/")[-1]
            if _board.fs_is_dir(f):
                print(f"{f_name}")
            else:
                print(ANSIEC.FG.BRIGHT_BLUE + f_name + ANSIEC.OP.RESET)
    except BaseException:
        print("The path " + ANSIEC.FG.BRIGHT_RED + "does not exist." + ANSIEC.OP.RESET)
                
def show_waiting(remote_filename, total_size):
    copied_size = 0
    bar_length = 40

    print(ANSIEC.FG.BRIGHT_BLUE + remote_filename.replace(_root_fs, "", 1) + ANSIEC.OP.RESET, flush=True)
    
    if total_size == 0:
        return
    
    while True:
        progress = min(copied_size / total_size, 1.0)    
        block = int(round(bar_length * progress))
        bar = "#" * block + "-" * (bar_length - block)
        print(ANSIEC.OP.left() + f"[{bar}] {int(progress * 100)}%", end="", flush=True)
        if progress >= 1.0:
            break
        time.sleep(0.1)
        if _type == 'xbee3':
            copied_size += (115200 // 8 // 100) * 0.8
        elif _type == 'pico2':
            copied_size += (115200 // 8 // 100) * 2
                    
    print(flush=True)

@upyboard.command()
@click.argument("local", type=click.Path(exists=True))
@click.argument("remote", required=False)
def put(local, remote):
    if remote is None:
        remote = os.path.basename(os.path.abspath(local))
    else:
        if not remote.startswith(_root_fs):
            remote = _root_fs + remote
        
        try:
            if _board.fs_is_dir(remote):
                remote = remote + "/" + os.path.basename(os.path.abspath(local))
        except BaseException:
            pass
        
    if os.path.isdir(local):
        _board.fs_putdir(local, remote, show_waiting)
    else:
        with open(local, "rb") as infile:        
            _board.fs_put(infile.read(), remote, show_waiting)

def _run_error_process(out, local_file):
    print(f"{'-' * 20} Traceback {'-' * 20}")
    
    for l in out[1:-2]:
        print(l.strip())
    
    try:
        err_line_raw = out[-2].strip()
        
        if "<stdin>" in err_line_raw:
            full_path = os.path.abspath(os.path.join(os.getcwd(), local_file))
            err_line = err_line_raw.replace("<stdin>", full_path, 1)
        else:
            match = re.search(r'File "([^"]+)"', err_line_raw)
            if match:
                full_path =  os.path.join(_board_lib_path, match.group(1))
                escaped_filename = re.sub(r"([\\\\])", r"\\\1", full_path)
                err_line = re.sub(r'File "([^"]+)"', rf'File "{escaped_filename}"', err_line_raw)
                
        print(f" {err_line}")
        
        err_content = out[-1].strip()

        match = re.search(r"line (\d+)", err_line)
        if match:
            line = int(match.group(1))
            try:
                with open(full_path, "r") as f:
                    lines = f.readlines()
                    print(f"  {lines[line - 1].rstrip()}")
            except:
                pass    

    except IndexError:
       err_content = out[-1].strip()
    

    print(ANSIEC.FG.BRIGHT_MAGENTA + err_content + ANSIEC.OP.RESET)
    
    
@upyboard.command()
@click.argument("local_file")
@click.option(
    "--no-waiting",
    "-n",
    is_flag=True,
    help="Do not join input/output stream",
)
@click.option(
    "--input-echo-on",
    "-i",
    is_flag=True,
    help="Turn on echo for input",
)
def run(local_file, no_waiting, input_echo_on):
    try:
        _board.run(local_file, not no_waiting, input_echo_on)
    except IOError:
        click.echo(f"File not found: {ANSIEC.FG.BRIGHT_RED + local_file + ANSIEC.OP.RESET}", err=True)
    except BaseException as ex:
        _run_error_process(str(ex).strip().split('\n'), local_file)
        
@upyboard.command()
def repl():
    print(ANSIEC.FG.MAGENTA + "Entering REPL mode. Press Ctrl + X to exit." + ANSIEC.OP.RESET)

    _board.repl()
    
@upyboard.command()
def format():
    print("Formatting...")
    ret = _board.fs_format(_type)
    if ret:
        print(ANSIEC.OP.left() + "Formatting is complete!")
    else:
        print(ANSIEC.OP.left() + "The device type is " + ANSIEC.FG.BRIGHT_RED + "not supported." + ANSIEC.OP.RESET)
    return ret


import mpy_cross

_base_src_path = None
_bytecode = None

def _mpy_output_path(filepath):
    relative_path = os.path.relpath(filepath, _base_src_path)
    output_dir = os.path.join(os.path.dirname(_base_src_path), "mpy", os.path.dirname(relative_path))
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_filename = os.path.splitext(os.path.basename(filepath))[0] + ".mpy"
    output_path = os.path.join(output_dir, output_filename)

    return output_path    

def _conv_to_mpy(dir):    
    for filename in os.listdir(dir):
        filepath = os.path.join(dir, filename)

        if os.path.isfile(filepath):
            outpath = _mpy_output_path(filepath)            
            if _bytecode == "v5":
                args = [filepath, '-o', outpath, '-mno-unicode', '-msmall-int-bits=31']
            else:
                args = [filepath, '-o', outpath, '-msmall-int-bits=31']
                        
            mpy_cross.run(*args)
                        
        elif os.path.isdir(filepath): 
            _conv_to_mpy(filepath) 

@upyboard.command()
def init():    
    global _base_src_path, _bytecode, _board
    
    _board.serial.close()
    
    for i in range(2):
        descript = is_micropython_board(_sport)   
        if descript:
            try:
                upy_version = float(re.search(r'v\d+\.\d+', descript).group(0)[1:])
                if upy_version >= 1.23:
                    _bytecode = "v6.3"
                elif upy_version >= 1.22:
                    _bytecode = "v6.2"
                elif upy_version >= 1.20:
                    _bytecode = "v6.1"
                elif upy_version >= 1.19:
                    _bytecode = "v6"
                elif upy_version >= 1.12:
                    _bytecode = "v5"
                else:
                    print("Unkown micropython version " + ANSIEC.FG.BRIGHT_RED + f"{descript}" + ANSIEC.OP.RESET)
                    raise click.Abort()
                break
            except AttributeError:
                print("Unkown Device " + ANSIEC.FG.BRIGHT_RED + f"{descript}" + ANSIEC.OP.RESET)
                raise click.Abort()
        else:
            if i > 0:                
                print("The device is not connected.", _sport)
                raise click.Abort()

    _board = Board(_sport, 115200)
    _base_src_path = os.path.join(_board_lib_path, "src")

    _base_mpy_path = os.path.join(_board_lib_path, "mpy")    
    if os.path.exists(_base_mpy_path):
        shutil.rmtree(_base_mpy_path)
        
    _pycache_path = os.path.join(_base_src_path, "__pycache__")
    if os.path.exists(_pycache_path):
        shutil.rmtree(_pycache_path)

    _conv_to_mpy(_base_src_path)
    
    if not click.Context(format).invoke(format):
        return 
    
    lib_root = _root_fs + "lib/"

    if _type == 'xbee3':
        remote = lib_root + "xnode"
    else:
        remote = lib_root

    local = os.path.join(_board_lib_path, "mpy")

    print("Installing the library on the board.")
        
    _board.fs_mkdir(lib_root)
    click.Context(put).invoke(put, local=local, remote=remote)
    
    if os.path.exists(_base_mpy_path):
        shutil.rmtree(_base_mpy_path)

    print("The job is done!")

@upyboard.command()
@click.argument("board", default="xnode")
def env(board):
    pass

@upyboard.command()
def scan():
    pass
#--------------------------------------------------------------
    
def main():
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])

    if len(sys.argv) == 2 and sys.argv[1] == "scan":
        device_scan()
    elif len(sys.argv) == 2 and sys.argv[1] == "env":
        host_environment("xnode")
    elif len(sys.argv) == 2 and sys.argv[1] == "--version":
        print("upyboard ", __version__)
    elif len(sys.argv) == 3 and sys.argv[1] == "env":
        board = sys.argv[2].lower().strip()
        print(board)
        host_environment(board)
    else:    
        if not any(item in sys.argv for item in ('get', 'put', 'rm', 'run')) and sys.argv[-1].split('.')[-1] == 'py':
            index = next((i for i, arg in enumerate(sys.argv[1:], 1) if arg in ['-i', '--input-echo-on', '-n', '--no-waiting']), None)
            if index is not None:
                sys.argv.insert(index, 'run')
            else:
                sys.argv.insert(-1, 'run')

        exit_code = upyboard()
        sys.exit(exit_code)
	
if __name__ == '__main__':
    main()


import utime
import uos
import machine


class ANSIEC:
    """
    Class that generates ANSI Escape Codes strings 
    """
    
    class FG:
        """
        Foreground Colors
        """
        BLACK = "\u001b[30m"
        RED = "\u001b[31m"
        GREEN = "\u001b[32m"
        YELLOW = "\u001b[33m"
        BLUE = "\u001b[34m"
        MAGENTA = "\u001b[35m"
        CYAN = "\u001b[36m"
        WHITE = "\u001b[37m"
        BRIGHT_BLACK= "\u001b[30;1m"
        BRIGHT_RED = "\u001b[31;1m"
        BRIGHT_GREEN = "\u001b[32;1m"
        BRIGHT_YELLOW = "\u001b[33;1m"
        BRIGHT_BLUE = "\u001b[34;1m"
        BRIGHT_MAGENTA = "\u001b[35;1m"
        BRIGHT_CYAN = "\u001b[36;1m"
        BRIGHT_WHITE = "\u001b[37;1m"
                
        @classmethod
        def rgb(cls, r:int, g:int, b:int) -> str: 
            """
            Returns ANSI code for foreground color as rgb value
            
            :param r: red value
            :param g: green value
            :param b: blue value
            
            :return: Returns ANSI code for foreground color
            """

    class BG:
        """
        Background Colors
        """
        BLACK = "\u001b[40m"
        RED = "\u001b[41m"
        GREEN = "\u001b[42m"
        YELLOW = "\u001b[43m"
        BLUE = "\u001b[44m"
        MAGENTA = "\u001b[45m"
        CYAN = "\u001b[46m"
        WHITE = "\u001b[47m"
        BRIGHT_BLACK= "\u001b[40;1m"
        BRIGHT_RED = "\u001b[41;1m"
        BRIGHT_GREEN = "\u001b[42;1m"
        BRIGHT_YELLOW = "\u001b[43;1m"
        BRIGHT_BLUE = "\u001b[44;1m"
        BRIGHT_MAGENTA = "\u001b[45;1m"
        BRIGHT_CYAN = "\u001b[46;1m"
        BRIGHT_WHITE = "\u001b[47;1m"
                
        @classmethod
        def rgb(cls, r:int, g:int, b:int) -> str:
            """
            Returns ANSI code for background color as rgb value
            
            :param r: red value
            :param g: green value
            :param b: blue value
            
            :return: Returns ANSI code for background color
            """

    class OP:
        """
        Operations
        """
        RESET = "\u001b[0m"
        BOLD = "\u001b[1m"
        UNDER_LINE = "\u001b[4m"
        REVERSE = "\u001b[7m"
        CLEAR = "\u001b[2J"
        CLEAR_LINE = "\u001b[2K"
        TOP = "\u001b[0;0H"

        @classmethod
        def up(cls, n:int) -> str:
            """
            Move cursor up
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor up
            """

        @classmethod
        def down(cls, n:int) -> str:
            """
            Move cursor down
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor down
            """
            
        @classmethod
        def right(cls, n:int) -> str:
            """
            Move cursor right
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor right
            """

        @classmethod
        def left(cls, n:int) -> str:
            """
            Move cursor left
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor left
            """
           
        @classmethod
        def next_line(cls, n:int) -> str:
            """
            Move cursor to the next line
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor to the next line
            """

        @classmethod
        def prev_line(cls, n:int) -> str:
            """
            Move cursor to the previous line
            
            :param n: Number of moves
            
            :return: Returns the ANSI Escape code to move the cursor to the previous line
            """
        
        @classmethod
        def to(cls, row, colum) -> str:
            """
            Move the cursor to that location
            
            :param row: Number of rows moved
            :param colum: Number of columns moved
            
            :return: Returns the ANSI Escape code to move the cursor
            """
            
def sqrt(x:int, epsilon:float=1e-10) -> float:
    """
    Newton's method for square root
    
    :param x: number to find square root
    :param epsilon: tolerance
    
    :return: square root of x
    """

def abs(x:int) -> int:
    """
    Absolute value
    
    :param x: number
    
    :return: absolute value of x
    """

def rand(size: int=4) -> int:
    """
    Generate random number
    
    :param size: size of random number
    
    :return: random number
    """

def map(x:float, min_i:float, max_i:float, min_o:float, max_o:float) -> float:
    """
    Map a number from one range to another
    
    :param x: number to map
    :param min_i: minimum value of input range
    :param max_i: maximum value of input range
    :param min_o: minimum value of output range
    :param max_o: maximum value of output range
    
    :return: mapped number
    """

def intervalChecker(interval:int) -> function:
    """
    Non-blogging interval checker
    
    :param interval: interval in milliseconds
    
    :return: function to check interval
    """

def WDT(timeout:int) -> machine.WDT:
    """
    Watchdog timer
        
    :param timeout: timeout in milliseconds
    """

def i2cdetect(bus:int=1, decorated:bool=True) -> str | list:
    """
    I2C device scanner
    
    :param bus: I2C bus number
    :param decorated: decorated output
    
    :return: I2C devices. 
    If decorated parameter is True, returns a string, otherwise returns a list.
    """
 
class Slip:
    """
    Class that handles SLIP protocol
    """

    @staticmethod
    def decode(chunk: bytes) -> list:
        """
        SLIP decoder. Returns a list of decoded byte strings.
        
        :param chunk: A byte string to decode.
        
        :return: A list of bytes.
        """
        
    @staticmethod
    def encode(payload: bytes) -> bytes:
        """
        SLIP encoder. Returns a byte string.
        
        :param payload: A byte string to encode.
        
        :return: A byte string.
        """
 
class ReplUart:
    """
    Class that handles UART communication
    """
    
    @property
    def timeout(self) -> None | int:
        """
        Get timeout
        
        :return: timeout
        """

    @timeout.setter
    def timeout(self, n:None|int):
        """
        Set timeout
        
        :param n: timeout
        """
        
    def read(self, size:int=1) -> bytes:
        """
        Read data from UART
        
        :param size: number of bytes to read. Default is 1
        
        :return: data
        """
        
    def read_until(self, expected:bytes=b'\n', size:None|int=None) -> bytes:
        """
        Read data from UART until expected data is received
        
        :param expected: expected data
        :param size: number of bytes to read
        
        :return: data
        """
        
    def write(self, data:bytes) -> int:
        """
        Write data to UART
        
        :param data: data to write  
        
        :return: number of bytes written
        """


import sys
import time
import uos
import machine

from micropython import const


class ANSIEC:
    class FG:
        BLACK = "\u001b[30m"
        RED = "\u001b[31m"
        GREEN = "\u001b[32m"
        YELLOW = "\u001b[33m"
        BLUE = "\u001b[34m"
        MAGENTA = "\u001b[35m"
        CYAN = "\u001b[36m"
        WHITE = "\u001b[37m"
        BRIGHT_BLACK= "\u001b[30;1m"
        BRIGHT_RED = "\u001b[31;1m"
        BRIGHT_GREEN = "\u001b[32;1m"
        BRIGHT_YELLOW = "\u001b[33;1m"
        BRIGHT_BLUE = "\u001b[34;1m"
        BRIGHT_MAGENTA = "\u001b[35;1m"
        BRIGHT_CYAN = "\u001b[36;1m"
        BRIGHT_WHITE = "\u001b[37;1m"
                
        @classmethod
        def rgb(cls, r, g, b): return "\u001b[38;2;{};{};{}m".format(r, g, b)

    class BG:
        BLACK = "\u001b[40m"
        RED = "\u001b[41m"
        GREEN = "\u001b[42m"
        YELLOW = "\u001b[43m"
        BLUE = "\u001b[44m"
        MAGENTA = "\u001b[45m"
        CYAN = "\u001b[46m"
        WHITE = "\u001b[47m"
        BRIGHT_BLACK= "\u001b[40;1m"
        BRIGHT_RED = "\u001b[41;1m"
        BRIGHT_GREEN = "\u001b[42;1m"
        BRIGHT_YELLOW = "\u001b[43;1m"
        BRIGHT_BLUE = "\u001b[44;1m"
        BRIGHT_MAGENTA = "\u001b[45;1m"
        BRIGHT_CYAN = "\u001b[46;1m"
        BRIGHT_WHITE = "\u001b[47;1m"
                
        @classmethod
        def rgb(cls, r, g, b): return "\u001b[48;2;{};{};{}m".format(r, g, b)

    class OP:
        RESET = "\u001b[0m"
        BOLD = "\u001b[1m"
        UNDER_LINE = "\u001b[4m"
        REVERSE = "\u001b[7m"
        CLEAR = "\u001b[2J"
        CLEAR_LINE = "\u001b[2K"
        TOP = "\u001b[0;0H"

        @classmethod
        def up(cls, n):
            return "\u001b[{}A".format(n)

        @classmethod
        def down(cls, n):
            return "\u001b[{}B".format(n)

        @classmethod
        def right(cls, n):
            return "\u001b[{}C".format(n)

        @classmethod
        def left(cls, n):
            return "\u001b[{}D".format(n)
        
        @classmethod
        def next_line(cls, n):
            return "\u001b[{}E".format(n)

        @classmethod
        def prev_line(cls, n):
            return "\u001b[{}F".format(n)
                
        @classmethod
        def to(cls, row, colum):
            return "\u001b[{};{}H".format(row, colum)

def sqrt(x, epsilon=1e-10):
    guess = x / 2.0

    op_limit = 5
    while abs(guess * guess - x) > epsilon and op_limit:
        guess = (guess + x / guess) / 2.0
        op_limit -= 1

    return guess

def abs(x):
    return x if x >= 0 else -x

def rand(size=4):
    return int.from_bytes(uos.urandom(size), "big")

def map(x, min_i, max_i, min_o, max_o):
    return (x - min_i) * (max_o - min_o) / (max_i - min_i) + min_o

def intervalChecker(interval):
    current_tick = time.ticks_us()   
    
    def check_interval():
        nonlocal current_tick
        
        if time.ticks_diff(time.ticks_us(), current_tick) >= interval * 1000:
            current_tick = time.ticks_us()
            return True
        return False
    
    return check_interval

def WDT(timeout):
    return machine.WDT(0, timeout)

def i2cdetect(bus=1, decorated=True):
    i2c = machine.I2C(bus)
    devices = i2c.scan()

    if decorated:
        output = "     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f"
        for i in range(0, 8):
            output += ANSIEC.FG.YELLOW + "\n{:02x}:".format(i*16) + ANSIEC.OP.RESET
            for j in range(0, 16):
                address = i * 16 + j
                if address in devices:
                    output += " {:02x}".format(address)
                else:
                    output += " --"
    else:
        output = devices
    
    return output

END = b'\xC0'
ESC = b'\xDB'
ESC_END = b'\xDC'
ESC_ESC = b'\xDD'


class Slip:
    __decode_state = {'started': False, 'escaped': False, 'data': bytearray(), 'pending_end': False, 'junk': bytearray()}
    
    @staticmethod
    def decode(chunk: bytes) -> list:
        """
        SLIP decoder. Returns a list of decoded byte strings.
        :param chunk: A byte string to decode.
        :return: A list of bytes.
        """
        result = []
        data = Slip.__decode_state['data']
        junk = Slip.__decode_state['junk']
        started = Slip.__decode_state['started']
        escaped = Slip.__decode_state['escaped']
        pending_end = Slip.__decode_state['pending_end']

        for char in chunk:
            if escaped:
                if char == ord(ESC_END):
                    data.append(ord(END))
                elif char == ord(ESC_ESC):
                    data.append(ord(ESC))
                else:
                    data.clear()
                    started = False
                    pending_end = False
                    return []
                escaped = False
            elif char == ord(ESC):
                escaped = True
            elif char == ord(END):
                if pending_end:
                    if started:
                        result.append(bytes(data))
                        data.clear()
                    else:
                        junk.clear()
                    started = True
                    pending_end = False
                elif started:
                    result.append(bytes(data))
                    data.clear()
                    started = False
                    pending_end = True
                else:
                    started = True
                    pending_end = True
            else:
                if pending_end:
                    started = True
                    data.append(char)
                    pending_end = False
                elif started:
                    data.append(char)
                else:
                    junk.append(char)

        Slip.__decode_state['started'] = started
        Slip.__decode_state['escaped'] = escaped
        Slip.__decode_state['pending_end'] = pending_end

        return result
    
    @staticmethod
    def encode(payload: bytes) -> bytes:
        """
        SLIP encoder. Returns a byte string.
        :param payload: A byte string to encode.
        :return: A byte string.
        """
        return END + payload.replace(ESC, ESC + ESC_ESC).replace(END, ESC + ESC_END) + END


class ReplUart:
    def __init__(self, timeout=None):
        self.timeout = timeout
    
    @property
    def timeout(self):
        return self.__timeout
    
    @timeout.setter
    def timeout(self, n):
        self.__timeout = n
    
    def read(self, size=1):        
        if self.timeout is None:
            assert size > 0, "size must be greater than 0"
            return sys.stdin.buffer.read(size)
        elif self.timeout is not None and self.timeout == 0:
            return sys.stdin.buffer.read(-1)
        elif self.timeout is not None and self.timeout > 0:
            rx_buffer = b''
            t0 = time.ticks_ms()
            while time.ticks_diff(time.ticks_ms(), t0) / 1000 < self.timeout:
                b = sys.stdin.buffer.read(-1)
                if b:
                    rx_buffer += b
                    if len(rx_buffer) >= size:
                        break
            return rx_buffer
        
    def read_until(self, expected=b'\n', size=None):
        rx_buffer = bytearray()
        expected_len = len(expected)

        t0 = time.ticks_ms() if (self.timeout is not None and self.timeout > 0) else None 
        while True:
            if t0 is not None:
                ellipsis = time.ticks_diff(time.ticks_ms(), t0) / 1000
                if ellipsis >= self.timeout:
                    break  # Timeout

            try:
                b = sys.stdin.buffer.read(-1)
                if self.timeout is not None and self.timeout == 0:
                    return b if b else b''
            except Exception as e:
                return
                
            if not b: 
                continue

            rx_buffer.extend(b)

            if size is not None and len(rx_buffer) >= size:
                return bytes(rx_buffer[:size])

            if len(rx_buffer) >= expected_len and rx_buffer[-expected_len:] == expected:
                return bytes(rx_buffer)
        
        return bytes(rx_buffer) # Timeout occurred, return what's read so far
                    
    def write(self, data:bytes) -> int:
        assert isinstance(data, bytes), "data must be a byte type"
        
        ret = sys.stdout.write(data)
                
        return ret



# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Union

TRANSPORT_TCP: int = ...
TRANSPORT_UDP: int = ...

IDLE: int = ...
SENDING: int = ...
SUCCESS: int = ...


class _device_request(object):
    """
    File-like object containing the payload of the request. This class cannot
    be instantiated, it is returned from
    ``digi.cloud.device_request_receive()``.

    **Note**: This class is only available on XBee3 Cellular devices.
    """

    def read(self, size: Optional[int] = -1) -> bytes:
        """
        Reads the payload received from Digi Remote Manager and returns up to
        ``size`` bytes. This call can block.

        :param size: The maximum number of bytes to read. If this parameter is
            omitted, ``None``, or negative, data is read and returned until
            EOF is reached.

        :return: Up to ``size`` bytes of payload from the device request or an
            empty bytes object if the stream is already at EOF.
        """
        ...

    def readinto(self, b: Union[bytearray, memoryview]) -> int:
        """
        Reads the payload received from Digi Remote Manager into a
        pre-allocated, ``bytearray()`` object ``b``, and returns the number of
        bytes read. This call, like ``read()`` is a blocking call.

        :param b: ``bytearray()`` object where payload is saved.

        :return: The number of bytes read.
        """
        ...

    def write(self, b: bytes) -> int:
        """
        Writes a response back to Digi Remote Manager. After finishing writing
        a response, ``close()`` method should be called to complete the
        transaction.

        **Note**: All data from the request should have been read before
        issuing a write.

        :param b: A ``bytes()`` or ``bytearray()`` object containing the answer
            to send.

        :return: The number of bytes written (always equal to the length of
            ``b`` in bytes)
        """
        ...

    def close(self) -> None:
        """
        Required method to finish the device request transaction. Transaction
        is finished sending a response to Digi Remote Manager.

        **Note**: The request should not be written to or read after a close
        call.
        """
        ...


def device_request_receive() -> Optional[_device_request]:
    """
    Polls for a device request from your Digi Remote Manager account.

    The returned ``device_request`` object is a file-like object containing
    the payload of the request that can be read and answered using the
    ``write`` method.

    **Note**: This feature is only available on XBee3 Cellular devices.

    :return: A device request from Digi Remote Manager or ``None`` if there is
        no pending request.
    """
    ...


class DataPoints(object):
    """
    Class used to create and upload new data streams and data points to your
    Digi Remote Manager account.

    **Note**: This class is only available on XBee3 Cellular devices.
    """

    def __init__(self, transport: int = TRANSPORT_TCP) -> None:
        """
        Class constructor. Instantiates a new ``DataPoints`` object with the
        provided parameters.

        :param transport: The transport method used to deliver the data points.
            Acceptable values are:

            * ``digi.cloud.TRANSPORT_TCP`` (default transport)
            * ``digi.cloud.TRANSPORT_UDP``
        """
        ...

    def __del__(self) -> None:
        """
        Frees up the resources held by the ``DataPoints`` object, usually when
        you are seeing ``OSError`` exceptions with **ENOBUFS**. Ensure that
        your MicroPython application has no references to the object anymore
        so that it can be garbage-collected. Usually you can use the ``del``
        statement to do this.

        The ``del`` statement is not necessary if the ``DataPoints`` object
        goes out of scope, for example when the ``DataPoints`` object was
        created inside of a function.

        **Note**: ``DataPoints`` object will not be garbage-collected if
        another object holds a reference to the ``DataPoints`` object. Placing
        the ``DataPoints`` object inside a container such as a list, tuple, or
        dictionary will cause this.
        """
        ...

    def add(self, stream_name: str, value: Any, *, units: Optional[str],
            quality: Optional[int], description: Optional[str],
            location: Optional[tuple]) -> None:
        """
        Creates a new data point entry inside the ``DataPoints`` container.

        If any of the parameters values are of an inappropriate type (such as
        an integer for stream name), a ``TypeError`` or ``ValueError`` is
        raised indicating the problem.

        This method will raise an ``OSError`` with the error code ``ENOSPC``
        when there is not enough room to add the data point to the upload
        buffer. The amount of space each data point consumes in the buffer
        varies based on the length of the string value and how many of the
        optional parameters are specified.

        **Note**: ``DataPoints`` objects that use the ``TRANSPORT_UDP``
        transport are limited to one data point per ``DataPoints`` container.

        :param stream_name: The data stream name to which this data point is
            added.
        :param value: The value to assign to this data point. Currently the
            only supported types are **integer**, **float** and **string**.
        :param units: A string, specifying the units associated with data on
            this data stream. If this value is specified, it will overwrite
            the units field of the data stream in Digi Remote Manager.
            Individual data points do not have units associated with them.
        :param quality: A user-defined 32-bit integer value indicating the
            quality of the data in the data point.
        :param description: A string, specifying a description of the data
            point.
        :param location: A tuple of three floating point numbers, indicating
            the geo-location information of the data point. Geo-location is
            represented as (latitude in degrees, longitude in degrees,
            elevation in meters).
        """
        ...

    def send(self, timeout: int = 30) -> None:
        """
        Performs a data point upload of all data that has been created using
        the ``add()`` method. This method blocks until the data has been
        uploaded or the specified timeout expires.

        * If the Digi Remote Manager feature is disabled (bit 0 of **ATDO** is
          cleared), this raises a ``TypeError`` indicating that the Remote
          Manager feature is disabled.
        * If there is no data to be uploaded, an ``OSError EINVAL`` is raised.
        * If a blocking upload fails (due to a network issue or command
          timeout), an ``OSError`` is raised.
        * If persistent TCP connections to Digi Remote Manager are disabled
          (**ATMO** bit 0 is cleared) and the transport selected is
          ``TRANSPORT_TCP``, this method causes a temporary TCP/SSL connection
          to be created.

        :param timeout: Number of seconds that the send call is allowed to
            block. If nonzero and this timeout elapsed without the data being
            sent, an ``OSError ETIMEDOUT`` is raised, but the object is still
            considered to be "locked" and new data points cannot be added.
        """
        ...

    def status(self) -> int:
        """
        Returns a value indicating the status of the most recent send call on
        a ``DataPoints`` object. Possible values are:

        * ``digi.cloud.IDLE``: Send has never been called.
        * ``digi.cloud.SENDING``: The most recent send call is still being
          processed.
        * ``digi.cloud.SUCCESS``: The most recent send call has succeeded.
        * Any other value is a negative ``uerrno`` value for the most recent
          ``send`` call. For example, ``uerrno.EIO``.

        :return: The status value of the most recent send call.
        """
        ...


class Console:
    """Class providing access to DRM console interface."""

    def isconnected() -> bool:
        """Indicates whether a console session is attached from DRM"""
        ...

    def read(self, size: int=-1) -> bytes:
        """
        Read from the DRM console.

        Reads up to ``size`` bytes from the object and returns them. As a
        convenience, if ``size`` is unspecified or ``-1``, all bytes until
        end-of-file (EOF) are returned.

        :param size: Maximum number of bytes to read.

        :return: Data from console client or None if no data is available.
        """
        ...

    def readinto(self, b: Union[bytearray, memoryview]) -> int:
        """
        Read from the DRM console into a pre-allocated buffer.

        Reads bytes into a pre-allocated, writable bytes-like object ``b``,
        and returns the number of bytes read. This call is non-blocking.

        :param b: Writable bytes-like object where read bytes will be placed.

        :return: Number of bytes read or None if no data is available.
        """
        ...
    def write(self, b: bytes) -> int:
        """Write to the DRM console.

        If a DRM console session is not attached, the data will be
        discarded.

        :param b: The value to send to the Console client in DRM.

        :return: The number of bytes written.

        """
        ...

    def close(self) -> None:
        """Close the current DRM console session."""
        ...



# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any


def hexlify(data: Any, sep: Any="") -> bytes:
    """
    Returns the hexadecimal representation of the provided binary data.

    :param data: Binary data to convert.
    :param sep: If supplied, this parameter is used as separator between
        hexadecimal values.

    :return: Bytes string with the hexadecimal representation.
    """
    ...

def unhexlify(data: Any) -> bytes:
    """
    Converts hexadecimal data to binary representation. Inverse of
    ``ubinascii.hexlify()``.

    :param data: Hexadecimal data to convert.

    :return: Bytes string with the binary representation.
    """
    ...

def a2b_base64(data: Any) -> bytes:
    """
    Decodes base64-encoded data, ignoring invalid characters in the input and
    returns the decoded data. Conforms to RFC 2045 s.6.8.

    :param data: The base64-encoded data to decode.

    :return: Bytes string with the decoded data.
    """
    ...

def b2a_base64(data: Any) -> bytes:
    """
    Encodes binary data in base64 format, as in RFC 3548 and returns the
    encoded data followed by a newline character, as a bytes object.

    :param data: Binary data to encode in base64 format.

    :return: Bytes string with the encoded data.
    """
    ...

def crc32(data: Any, crc: int=0) -> int:
    """
    Computes CRC-32 incrementally, the 32-bit checksum of ``data``, starting
    with an initial CRC of ``crc``.

    :param data: Data to obtain its 32-bit checksum.
    :param crc: Starting CRC value.

    :return: 32-bit checksum of the provided data.
    """
    ...



# Copyright (c) 2019-2020, Digi International Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides access to Bluetooth Low Energy functionality.

**Note**: Unless otherwise specified, the ``digi.ble`` module
and all of its contents are new in the following firmware versions:
  * XBee3 Zigbee: version 1009
  * XBee3 802.15.4: version 200A
  * XBee3 DigiMesh 2.4: version 300A
  * XBee3 Cellular LTE-M/NB-IoT: version 11415
  * XBee3 Cellular LTE Cat 1: version x15
"""

from typing import (
    Any, Callable, ContextManager, Iterable, Iterator, List, Optional, NewType,
    Sized, Tuple, Union, overload
)

try:
    # Python 3.8 (and PyCharm 2019.2.5 and newer) support TypedDict,
    # which means the _GAPScanDict type can actually provide useful hints.
    # (PyCharm issue PY-36008.)
    # In case the PyCharm installation being used does not have TypedDict defined,
    # fall back to using just a regular dict type.

    from typing import TypedDict
    class _GAPScanDict(TypedDict):
        address: bytes
        addr_type: int
        connectable: bool
        rssi: int
        payload: bytes
except ImportError:
    from typing import Dict
    _GAPScanDict = Dict

_UUIDValue = Union[int, str, bytes]
_UUIDOrValue = Union['UUID', _UUIDValue]

# These type names are provided mainly for documentation purposes,
# i.e. to make the PyCharm documentation popup use Service, Characteristic,
# etc. instead of just int.
# However, they also have a secondary benefit, which is that a type checker
# can flag instances where a service handle is passed as a characteristic
# or descriptor, or vice versa.
_ServiceHandle = NewType('Service', int)
_CharacteristicHandle = NewType('Characteristic', int)
_DescriptorHandle = NewType('Descriptor', int)
_Properties = NewType('Properties', int)

_ServiceTuple = Tuple[_ServiceHandle, 'UUID']
_CharacteristicTuple = Tuple[_CharacteristicHandle, 'UUID', _Properties]
_DescriptorTuple = Tuple[_DescriptorHandle, 'UUID']


__all__ = (
    'ADDR_TYPE_PUBLIC', 'ADDR_TYPE_RANDOM', 'ADDR_TYPE_PUBLIC_IDENTITY',
    'ADDR_TYPE_RANDOM_IDENTITY', 'ADDR_TYPE_STATIC', 'ADDR_TYPE_RANDOM_RESOLVABLE',
    'ADDR_TYPE_RANDOM_NONRESOLVABLE', 'PROP_BROADCAST', 'PROP_READ',
    'PROP_WRITE', 'PROP_WRITE_NO_RESP', 'PROP_AUTH_SIGNED_WR',
    'PROP_NOTIFY', 'PROP_INDICATE', 'PAIRING_REQUIRE_MITM',
    'PAIRING_REQUIRE_BONDING', 'PAIRING_DISABLE_LEGACY',
    'BLE_1M_PHY', 'BLE_2M_PHY', 'BLE_125K_CODED_PHY', 'BLE_500K_CODED_PHY',
    'active', 'config', 'gap_advertise', 'gap_scan', 'gap_connect',
    'xbee_connect', 'UUID',
    'format_address', 'parse_address',
)


ADDR_TYPE_PUBLIC: int = ...
ADDR_TYPE_RANDOM: int = ...
ADDR_TYPE_PUBLIC_IDENTITY: int = ...
ADDR_TYPE_RANDOM_IDENTITY: int = ...

ADDR_TYPE_STATIC: int = ...
ADDR_TYPE_RANDOM_RESOLVABLE: int = ...
ADDR_TYPE_RANDOM_NONRESOLVABLE: int = ...

PROP_BROADCAST: int = ...
PROP_READ: int = ...
PROP_WRITE: int = ...
PROP_WRITE_NO_RESP: int = ...
PROP_AUTH_SIGNED_WR: int = ...
PROP_NOTIFY: int = ...
PROP_INDICATE: int = ...

PAIRING_REQUIRE_MITM: int = ...
PAIRING_REQUIRE_BONDING: int = ...
PAIRING_DISABLE_LEGACY: int = ...

BLE_1M_PHY: int = ...
BLE_2M_PHY: int = ...
BLE_125K_CODED_PHY: int = ...
BLE_500K_CODED_PHY: int = ...


def active(
        active: Optional[bool] = None,
        /,  # active can only be specified as a positional argument
) -> bool:
    """
    Sets or queries whether Bluetooth Low Energy (BLE) functionality
    is enabled on the XBee3 module.

    :param active:
        If True, enable BLE functionality.
        If False, disable BLE functionality.
        If not given, query whether BLE functionality is currently enabled.
    :return: True if BLE functionality is enabled, False otherwise.
    """
    ...


def config(
        name: Optional[str] = None,
        /,  # name can only be specified as a positional argument
        *,  # settings can only be changed using keyword arguments
        interval_ms: int = -1,
        latency: int = -1,
        timeout_ms: int = -1,
) -> Any:  # at this time: bytes, int or None
    """
    Query a BLE configuration value by name. Or, update one or more BLE
    configuration values by name.

    **Query a configuration option**

    To query a BLE configuration option, pass the name of the option.
    The currently supported option names are:

        * ``"mac"``: Get the device BLE MAC address as a ``bytes`` object.
          (This is equivalent to querying the ATBL AT command.)
        * ``"interval_ms"``: See the ``interval_ms`` parameter below.
        * ``"latency"``: See the ``latency`` parameter below.
        * ``"timeout_ms"``: See the ``timeout_ms`` parameter below.

    **Update configuration options**

    To change the value of one or more BLE configuration options,
    call ``config()`` using keyword arguments. For example::

        ble.config(interval_ms=100, timeout_ms=5000)

    When modifying one or more configuration options, other options which
    are not specified will not be modified (i.e. their current values will
    be left as is).

    **Note:** XBee3 Zigbee firmware version 1009 only supports
    ``config("mac")``. Support for other query values and keyword
    arguments was added in firmware 100A.

    :param name: Name of a BLE configuration value to query.
        When querying a value, settings (using keyword arguments) are not
        allowed.
    :param interval_ms: The default connection interval (the time between two
        data transfer events) for future GAP connections.
        The value will be rounded down to the nearest multiple of 1.25 milliseconds.
        ``interval_ms`` may be between 8 and 4000 (4 seconds).
        The default interval value (restored at XBee power-up) is 50 milliseconds.
    :param latency: The default slave latency value for future GAP connections.
        Slave latency is the number of connection events which the peripheral
        is allowed to skip before the connection is dropped.
        ``latency`` may be between 0 and 500.
        The default latency (restored at XBee power-up) is 0.
    :param timeout_ms: The default connection supervision timeout value for
        future GAP connections.  The supervision timeout is the time that the
        central device (the XBee, in this case) wil wait for a data transfer
        before assuming that the connection is lost.
        ``timeout_ms`` may be between 100 and 32000 (32 seconds).
        ``timeout_ms`` must be larger than 2 * interval_ms * (latency + 1).
        The default timeout value (restored at XBee power-up) is 1 second.
    :return: The queried configuration value, or None if setting.
    :raises ValueError: One or more parameters was invalid.
    :raises TypeError: Improper number or type of arguments. For example,
        positional and keyword arguments cannot be mixed (i.e. you cannot
        specify a name as a string while applying a setting).
    :raises OSError: New configuration could not be applied.
    """
    ...


def format_address(address: bytes, /) -> str:
    r"""
    Convert a BLE 48-bit MAC address (6 bytes) into a human-readable string.

    The ``"address"`` values in advertisement data collected with ``gap_scan``
    are an example of suitable inputs to this function.

    Example::

        ble.format_address(b'\x01\x25\xb7\x93\x84\x6f')
        # "01:25:B7:93:84:6F"

    **Note:**

    The ``format_address`` function is new in the following firmware versions:
        * XBee3 Cellular: version x1C

    :param address: ``bytes`` object of length 6, representing the raw 48-bit
                    BLE MAC address of a device. Note that ``format_address``
                    can only take this parameter as a positional argument,
                    not a keyword argument.
    :return: a ``str``, formatted as ``"XX:XX:XX:XX:XX:XX"``
    :raises TypeError: ``address`` is not of the correct type
    :raises ValueError: ``address`` is of the wrong length
    """
    ...


def parse_address(address: str, /) -> bytes:
    r"""
    Convert a human-readable BLE MAC address string into a 48-bit
    (6 byte) value.

    The resulting ``bytes`` value is suitable for use in a call to
    ``gap_connect``.

    Example::

        address = ble.parse_address("01:25:B7:93:84:6F")
        conn = ble.gap_connect(ble.ADDR_TYPE_PUBLIC, address)

    **Note:**

    The ``parse_address`` function is new in the following firmware versions:
        * XBee3 Cellular: version x1C

    :param address: ``str`` object, consisting of 12 hexadecimal digits and
                    up to 5 colon (``:``) characters. Note that
                    ``parse_address`` can only take this parameter as a
                    positional argument, not a keyword argument.
    :return: a ``bytes``, with the 48-bit form of ``address``
    :raises TypeError: ``address`` is not of the correct type
    :raises ValueError: ``address`` is too short, too long, or does not contain
                        exactly 12 hexadecimal digits
    """
    ...


def gap_advertise(
        interval_us: Optional[int],
        /,  # interval_us can only be specified as a positional argument
        adv_data: Optional[bytes] = None
) -> None:
    r"""
    Start or stop GAP advertisements from the XBee3 module.

    **Stop advertising:**

    To stop advertising, set ``interval_us`` to ``None``. Example::

        from digi import ble
        ble.gap_advertise(None)

    **Return to default advertisement:**

    If ``adv_data`` is empty (i.e. ``b''``), then GAP advertising will return
    to the default XBee behavior, which is to advertise the product name
    (e.g. ``XBee3 Zigbee``) and ``interval_us`` is ignored.::

        ble.gap_advertise(500000, b'')

    **Specify new advertisement data:**

    If ``adv_data`` is given and is not ``None`` or empty (i.e. ``b''``),
    then it should consist of one or more Advertising Data (AD) elements,
    as defined in the Bluetooth Core Specification Supplement, Part A Section 1.
    The format of Advertising Data elements is summarized here:
      * Each AD element consists of a length byte, a data type byte, and
        one or more bytes of data. The length byte indicates how long the rest
        of the element is, e.g. a Complete Local Name element with the value
        ``"My XBee"`` would have a length byte ``0x08`` - 1 byte for type plus
        7 bytes for the value.
      * The Bluetooth SIG provides the list of defined Advertising Data element types
        at https://www.bluetooth.com/specifications/assigned-numbers/generic-access-profile/

    Please be aware that in order to advertise a specific device name (such as ``"My XBee"``),
    ``adv_data`` cannot consist simply of the name to be advertised.
    The ``adv_data`` value *must* be formatted as one or more Advertising Data elements
    in order to be interpreted as a valid Bluetooth Low Energy advertisement by other devices.

    Example::

        ble.gap_advertise(200000, b"\x08\x09My XBee")

        # Example using a bytearray
        name = b"My XBee"
        data = bytearray([len(name) + 1, 0x09])
        data.extend(name)
        ble.gap_advertise(200000, data)

    **Re-use data from the previous gap_advertise call:**

    If ``adv_data`` is ``None`` or not specified, then the data passed to the
    previous call to ``gap_advertise`` will be re-used, unless there was no
    previous call or the data was empty, in which case the behavior will be
    as if an empty value (``b''``) was passed.::

        # These two calls are equivalent.
        ble.gap_advertise(200000)
        ble.gap_advertise(200000, None)

    :param interval_us: Start advertising at the specified interval, in microseconds.
        This value will be rounded down to the nearest multiple of 625 microseconds.
        If this value is not ``None``, it must be at least 20,000 microseconds
        (20 milliseconds) and no larger than 40,959,375 microseconds
        (approximately 40.96 seconds).
        To stop advertising, set ``interval_us`` to ``None``.
    :param adv_data: The payload that will be included in GAP advertisement broadcasts.
        ``adv_data`` can be a ``bytes`` or ``bytearray`` object up to 31 bytes in length,
        or ``None``.
    :raises OSError ENODEV: BLE functionality is disabled (see ``active()``).
    """
    ...


class _gap_scan(ContextManager, Iterable[_GAPScanDict]):
    """
    Class used to encapsulate an ongoing GAP scan (device discovery) operation,
    This class cannot be instantiated, it is returned from ``digi.ble.gap_scan()``.

    See the documentation for ``digi.ble.gap_scan()`` for examples of usage.
    """

    def get(self) -> List[_GAPScanDict]:
        """
        Return a list of all received GAP advertisements which are currently in the
        internal queue. This list may be empty.

        If the GAP scan has timed out (see ``duration_ms`` argument to ``gap_scan``)
        or if ``stop()`` has been called, this method will return any remaining
        advertisements, but no new advertisements will be stored.

        Each advertisement is formatted as a dictionary containing the following fields:

        * ``address``: The BLE MAC address of the sender of the advertisement, formatted
          as a ``bytes`` object.
        * ``addr_type``: The type of address contained in the ``address`` field. The possible
          values are defined as constants on the ``digi.ble`` module:
            * ``ADDR_TYPE_PUBLIC``
            * ``ADDR_TYPE_RANDOM``
            * ``ADDR_TYPE_PUBLIC_IDENTITY``
            * ``ADDR_TYPE_PUBLIC_RANDOM``
        * ``connectable``: ``True`` if the advertising device indicates that BLE central-mode devices
          may connect to it, ``False`` otherwise.
        * ``rssi``: The received signal strength of the advertisement, in dBm.
        * ``payload``: The raw advertisement payload, formatted as a ``bytes`` object.

        :return: a list (possibly empty) of received GAP advertisements, where each
            advertisement is a dictionary whose fields are described above.
        """
        ...

    def any(self) -> bool:
        """
        Return a boolean indicating whether there are any GAP advertisements in the internal queue
        (i.e. whether ``get()`` would return any advertisements).

        :return: ``True`` if there are advertisements waiting to be accessed, ``False`` otherwise.
        """
        ...

    def stop(self) -> None:
        """
        Stop the ongoing GAP scan operation.
        """
        ...

    def stopped(self) -> bool:
        """
        Return a boolean indicating whether the GAP scan operation has been stopped
        (using ``stop()``) or has timed out (see ``duration_ms`` argument to ``gap_scan``).

        :return: ``True`` if the GAP scan operation has been stopped or has timed out,
            ``False`` otherwise
        """
        ...

    def __enter__(self) -> _gap_scan:
        """
        Enter the runtime context for using this GAP scan object as a context manager
        (using the ``with`` statement).

        This step has no effect, but the call to ``__exit__`` which occurs
        when this context is exited will have the same effect as ``stop()``.

        You do not need to call ``__enter__`` directly, this happens automatically
        when the ``with`` statement is used.

        Example::

            with ble.gap_scan(0) as scan:  # __enter__ is called
                for adv in scan:
                    print(adv)
                    # Break out of the loop...
                    break
                # ... since we do nothing here, the `with` context is exited.
            # __exit__ is called
            assert scan.stopped()

        :return: this GAP scan object
        """
        ...


def gap_scan(
        duration_ms: int = 0,
        interval_us: int = 1280000,
        window_us: int = 11250,
        *,  # oldest is keyword-only
        oldest: bool = False,
) -> _gap_scan:
    """
    Start a GAP scan (discovery) operation to collect advertisements from nearby BLE devices.

    Each advertisement is formatted as a dictionary containing the following fields:

    * ``address``: The BLE MAC address of the sender of the advertisement, formatted
      as a ``bytes`` object.
    * ``addr_type``: The type of address contained in the ``address`` field. The possible
      values are defined as constants on the ``digi.ble`` module:
        * ``ADDR_TYPE_PUBLIC``
        * ``ADDR_TYPE_RANDOM``
        * ``ADDR_TYPE_PUBLIC_IDENTITY``
        * ``ADDR_TYPE_PUBLIC_RANDOM``
    * ``connectable``: ``True`` if the advertising device indicates that BLE central-mode devices
      may connect to it, ``False`` otherwise. (Note: XBee3 modules do not currently support
      connecting to other BLE devices.)
    * ``rssi``: The received signal strength of the advertisement, in dBm.
    * ``payload``: The raw advertisement payload, formatted as a ``bytes`` object.

    **Accessing the received advertisements:**

    One way to access the advertisements received during a GAP scan operation
    is to call the ``get()`` method on the scan object, which returns a list
    of advertisements::

        scan = ble.gap_scan()
        # Pause for a few seconds to allow some advertisements to arrive.
        time.sleep(5)
        for advertisement in scan.get():
            print(advertisement)
        scan.stop()

    Continue reading for other examples of accessing the advertisements.

    **Using gap_scan as an iterator:**

    Instead of calling ``get()`` repeatedly to access received GAP advertisements,
    the object returned by ``gap_scan`` may be used as an iterator, or in other words,
    as the target of a ``for``-loop::

        scan = ble.gap_scan(duration_ms=10000)  # 10 seconds
        for advertisement in scan:
            print(advertisement)

    Using ``gap_scan`` as an iterator is the preferred means to access the
    received advertisements. This is because calling ``get()`` requires allocating a list
    and filling the list with advertisements (dictionaries), whereas using ``gap_scan``
    as an iterator is more efficient because only one dictionary needs to be created at a time.

    To stop iterating over the ``gap_scan`` object (in order for your MicroPython application
    to perform other activities, for example), just ``break`` out of the loop. The GAP scan
    will continue running in the background. Example::

        # Example: read and print out 5 advertisements at a time
        def process_five_advertisements(scan):
            count = 0
            for advertisement in scan:
                print(advertisement)
                count += 1
                if count >= 5:
                    break

        scan = ble.gap_scan()
        while True:
            process_five_advertisements()
            do_something_else()
        scan.stop()

    **Using gap_scan as a context manager:**

    When running an indefinite GAP scan operation (``duration_ms=0``), instead of
    needing to call ``stop()`` to end the GAP scan operation, you may instead use
    the object returned by ``gap_scan`` as a context manager. By doing this,
    when the ``with`` block is exited, the GAP scan operation is automatically stopped.
    This approach uses less code and is less error-prone.

    For example, if you want to run a GAP scan operation until any advertisement whose
    payload contains a particular byte string is found, you can do this as follows::

        def find_advertisement(search):
            with ble.gap_scan(0) as scan:
                for adv in scan:
                    if search in adv["payload"]:
                        return adv

        found = find_advertisement(b"Hello, XBee")

    Compare to this example which does not use a context manager. Note the use of a try/finally block,
    and the need to call ``stop()`` explicitly::

        def find_advertisement(search):
            scan = ble.gap_scan(0)
            try:
                for adv in scan:
                    if search in adv["payload"]:
                        return adv
            finally:
                # Make sure to call stop() even if an exception is raised.
                scan.stop()

        found = find_advertisement(b"Hello, XBee")

    :param duration_ms: The duration of the GAP scan operation, i.e. how long to scan, in milliseconds.
        To scan indefinitely, set this to 0. If this parameter is not specified, it defaults
        to 0 (indefinite scan).
    :param interval_us: Optionally configure the duty cycle of the GAP scan operation.
        The scanner will run for ``window_us`` microseconds every ``interval_us`` microseconds.
        ``window_us`` must be less than or equal to ``interval_us``.
        ``interval_us`` must be at least 2,500 microseconds (2.5 milliseconds)
        and no more than approximately 40.96 seconds (40,959,375 microseconds).
        The default interval is 1.28 seconds.
    :param window_us: Optionally configure the duty cycle of the GAP scan operation.
        The scanner will run for ``window_us`` microseconds every ``interval_us`` microseconds.
        ``window_us`` must be less than or equal to ``interval_us``.
        ``window_us`` must be at least 2,500 microseconds (2.5 milliseconds)
        and no more than approximately 40.96 seconds (40,959,375 microseconds).
        The default window is 11.25 milliseconds.
    :param oldest: Discard new received advertisements once the internal queue fills up.
        Normally this should be left at its default value of False, so that new advertisements
        will overwrite the oldest stored advertisement when the queue fills up.
        Only set this to True if your application requires this behavior.
    :return: an object which encapsulates the GAP scan operation.
    :raises OSError ENODEV: BLE functionality is disabled (see ``active()``).
    :raises OSError EALREADY: Another ``gap_scan`` is still running.
        (Call ``.stop()`` on the existing scan. If the existing scan object
        cannot be referenced, soft reset the MicroPython REPL.)
    """
    ...


class UUID(Sized):
    """
    A representation of a Bluetooth UUID value.

    To read the UUID value, convert the ``UUID`` object into a ``bytes``
    or ``bytearray`` object::

        uuid = some_characteristic.uuid()
        uuid_value = bytes(uuid)

    The UUID value is either 16 bits (e.g. ``0x2908``)
    or 128 bits (e.g. ``7dddca00-3e05-4651-9254-44074792c590``).

    When a ``UUID`` object is converted into a string,
    by passing it to ``print`` or using ``str()`` or ``repr()``,
    the UUID value is presented in a human-readable format.
    This is useful for logging or printing messages to the console.
    For example::

        UUID(0x2908)
        UUID('7dddca00-3e05-4651-9254-44074792c590')

    To query the size (length in bytes) of the UUID value, use the built-in
    **len** function::

        len(ble.UUID(0x1234))  # returns 2
        len(ble.UUID('7dddca00-3e05-4651-9254-44074792c590'))  # returns 16

    **Note:**

    The ``UUID`` class is new in the following firmware versions:
        * XBee3 Zigbee: version 100A
        * XBee3 802.15.4: version 200A
        * XBee3 Cellular LTE-M/NB-IoT: version 11415
        * XBee3 Cellular LTE Cat 1: version x15
    """

    def __init__(self, uuid: _UUIDValue, /) -> None:
        """
        Create a new ``UUID`` object referencing the given UUID value.

        If ``uuid`` is an integer, it is treated as a 16-bit value.

        If ``uuid`` is a string or bytes of the form ``0xXXXX``
        (in other words, the hexadecimal form of a 16-bit value),
        it is converted into the equivalent 16-bit integer value.

        If ``uuid`` is a string or bytes of the form
        ``XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX``
        (in other words, the string form of a 128-bit UUID),
        it is converted into the equivalent 128-bit (16-byte) value.

        :raises ValueError: ``uuid`` is invalid
        :raises TypeError: ``uuid`` is of the incorrect type
        """
        ...


class _gap_connect(ContextManager):
    """
    Class used to encapsulate a GAP connection to a remote BLE peripheral device.
    This class cannot be instantiated, it is returned from ``digi.ble.gap_connect()``.

    See the documentation for ``digi.ble.gap_connect()`` for examples of usage.
    """

    def addr(self) -> Tuple[int, bytes]:
        """
        Returns the BLE peripheral device's address and address type.

        The returned address is a tuple consisting of two fields:
          * BLE address type of the peripheral device.
          * BLE address of the peripheral device, formatted as a ``bytes`` object.

        :return: A 2-tuple containing the BLE addressing information.
        """
        ...

    def close(self) -> None:
        """
        Close the GAP connection.

        The connection object will no longer be usable.
        """
        ...

    def config(
        self,
        name: Optional[str] = None,
        /,  # name can only be specified as a positional argument
        *,  # settings can only be changed using keyword arguments
        interval_ms: int = -1,
        latency: int = -1,
        timeout_ms: int = -1,
    ) -> Any:  # at this time: bytes, int or None
        """
        Query or update the BLE timing parameters for this connection.

        To control these timing parameters *before* opening a connection,
        see ``digi.ble.config()``.

        **Query a configuration option**

        To query a connection option, pass the name of the option.
        The currently supported option names are:

            * ``"interval_ms"``: See the ``interval_ms`` parameter below.
            * ``"latency"``: See the ``latency`` parameter below.
            * ``"mtu"``: Get the MTU value.
            * ``"timeout_ms"``: See the ``timeout_ms`` parameter below.

        **Update configuration options**

        To change the value of one or more BLE connection options,
        call ``config()`` using keyword arguments. For example::

            connection.config(interval_ms=100, timeout_ms=5000)

        When modifying one or more configuration options, other options which
        are not specified will not be modified (i.e. their current values will
        be left as is).

        :param name: Name of a BLE connection option to query.
            When querying a value, settings (using keyword arguments) are not
            allowed.
        :param interval_ms: The connection interval (the time between two
            data transfer events) for this GAP connection. The value will be
            rounded down to the nearest multiple of 1.25 milliseconds.
            ``interval_ms`` may be between 8 and 4000 (4 seconds).
            The default interval value (restored at XBee power-up) is
            50 milliseconds.
        :param latency: The slave latency value for this GAP connection.
            Slave latency is the number of connection events which the
            peripheral is allowed to skip before the connection is dropped.
            ``latency`` may be between 0 and 500.
            The default latency (restored at XBee power-up) is 0.
        :param timeout_ms: The connection supervision timeout value for this
            GAP connection. The supervision timeout is the time that the
            central device (the XBee, in this case) wil wait for a data transfer
            before assuming that the connection is lost.
            ``timeout_ms`` may be between 100 and 32000 (32 seconds).
            ``timeout_ms`` must be larger than 2 * interval_ms * (latency + 1).
            The default timeout value (restored at XBee power-up) is 1 second.
        :return: The queried configuration value, or None if setting.
        :raises ValueError: One or more parameters was invalid.
        :raises OSError: New configuration could not be applied (typically
            ENOTCONN, the connection was lost)
        """

    def phyconfig(
        self,
        preferred: int,
        accepted: int,
    ) -> int:
        """
        Update the BLE PHY used for this connection if it is also supported by
        the remote device.

        **Update PHY configuration**

        To attempt to change the PHY of the connection to the 2M PHY call ``phyconfig``
        and set the preferred phy. For example::

            connection.phyconfig(preferred=BLE_2M_PHY, accepted=BLE_2M_PHY)

        :param preferred: Used for specifying the PHYs that the local module
            will attempt to negotiate with the connected module. This field is a
            bitfield and can have one or more of BLE_1M_PHY, BLE_2M_PHY, BLE_125K_CODED_PHY
            or BLE_500K_CODED_PHY set.
        :param accepted: Used for specifying the PHYs that the module
            can accept in a remote initiated PHY update request. A PHY update will not
            occur if none of the accepted PHYs are present in the request. This field is a
            bitfield and can have one or more of BLE_1M_PHY, BLE_2M_PHY, BLE_CODED_PHY (0x4),
            or BLE_ANY_PHY (0xFF) set.
        :return: The phy the connection negotiated to use.
        :raises ValueError: One or more parameters was invalid.
        :raises OSError: New configuration could not be applied (typically
            ENOTCONN, the connection was lost)

        The ``phyconfig`` function is new in the following firmware versions:
            * XBee3 ZigBee: version 1014
            * XBee3 802.15.4: version 2014
            * XBee3 DigiMesh 2.4: version 3014
            * XBee3 Cellular: version x1C
        """
        ...

    def disconnect_code(self) -> int:
        """
        When called on a connection which has been closed, returns a value
        from the Bluetooth Core specification Vol 2, Part D (Error
        Codes) indicating the reason for the disconnect.  Calling this
        on an open connection returns zero.

        The most common values to see include:

        8 - Connection timeout
        19 - Remote user terminated
        22 - Connection terminated by local host

        """
        ...

    def gattc_services(
        self,
        uuid: Optional[_UUIDOrValue] = None,
    ) -> Iterator[_ServiceTuple]:
        """
        Perform a search/discovery of all GATT services in the remote device's
        GATT server database.

        Each returned service is a tuple consisting of three fields:
          * The service handle
          * The service's UUID value, as a ``UUID`` object

        Note that this function returns an iterator of service tuples,
        and the service discovery operation will not terminate until the
        iterator is emptied (all values are consumed). An error will be raised
        if another service, characteristic or descriptor discovery is started
        before the ongoing discovery completes.

        Example usage::

            with ble.gap_connect(ble.ADDR_TYPE_PUBLIC, remote_mac) as conn:
                for service in conn.gattc_services():
                    print(service)

                # Or, you can consume the items in the iterator into a list.
                services = list(conn.gattc_services())

        :param uuid: A ``UUID`` value or a value that can be used to construct a
            ``UUID`` (int, bytes or UUID string). If given and not None,
            the iterator only generates the service with the given UUID
            (if it exists), otherwise all discovered services are generated.
        :return: An iterator of service tuples, where each tuple contains a
            a service handle and a ``UUID`` value.
        :raises OSError: Service discovery operation could not be started.
        """
        ...

    def gattc_characteristics(
        self,
        service: Union[_ServiceHandle, _ServiceTuple],
        /,  # service can only be specified as a positional argument
        uuid: Optional[_UUIDOrValue] = None,
    ) -> Iterator[_CharacteristicTuple]:
        """
        Perform a search/discovery of all GATT characteristics on the given service
        in the remote device's GATT server database.

        ``service`` can either be a GATT service handle (one of the values contained
        in the tuple(s) generated by ``gattc_services()``), or one of the tuples
        generated by ``gattc_services()`` (in which case the service handle will be
        taken from the tuple).

        Each returned characteristic is a tuple consisting of three fields:
          * The characteristic handle
          * The characteristic's UUID value, as a ``UUID`` object
          * The characteristic's property flags, which is a bitmask of flags defined
            as constants on the ``digi.ble`` module:
              * ``PROP_BROADCAST``
              * ``PROP_READ``
              * ``PROP_WRITE``
              * ``PROP_WRITE_NO_RESP``
              * ``PROP_AUTH_SIGNED_WR``
              * ``PROP_NOTIFY``
              * ``PROP_INDICATE``

        Note that this function returns an iterator of characteristic tuples,
        and the characteristic discovery operation will not terminate until the
        iterator is emptied (all values are consumed). An error will be raised
        if another service, characteristic or descriptor discovery is started
        before the ongoing discovery completes.

        Example usage::

            with ble.gap_connect(ble.ADDR_TYPE_PUBLIC, remote_mac) as conn:
                # Since characteristic discovery cannot be started while
                # another discovery is in process, we must consume the services
                # into a list before discovering their characteristics.
                services = list(conn.gattc_services())

                for service in services:
                    for characteristic in conn.gattc_characteristics(service):
                        print(service, characteristic)

        :param service: A service handle, or one of the service tuples returned by
            ``gattc_services()``.
        :param uuid: A ``UUID`` value or a value that can be used to construct a
            ``UUID`` (int, bytes or UUID string). If given and not None,
            the iterator only generates characteristics found with the given UUID
            (if it exists), otherwise all discovered characteristics are generated.
        :return: An iterator of characteristic tuples, where each tuple contains
            a characteristic handle (an integer), a ``UUID`` value, and the
            characteristic's property flags.
        :raises OSError: Characteristic discovery operation could not be started.
        """
        ...

    def gattc_descriptors(
        self,
        characteristic: Union[_CharacteristicHandle, _CharacteristicTuple],
        /,  # positional argument only
    ) -> Iterator[_DescriptorTuple]:
        """
        Perform a search/discovery of all GATT descriptors on the given
        characteristic in the remote device's GATT server database.

        ``characteristic`` can either be a GATT characteristic handle
        (one of the values contained in the tuple(s) generated by
        ``gattc_characteristics()``), or one of the tuples generated by
        ``gattc_characteristics()`` (in which case the service handle will be
        taken from the tuple).

        Each returned descriptor is a tuple consisting of two fields:
          * The descriptor handle
          * The descriptor's UUID value, as a ``UUID`` object

        Note that this function returns an iterator of descriptor tuples,
        and the descriptor discovery operation will not terminate until the
        iterator is emptied (all values are consumed). An error will be raised
        if another service, characteristic or descriptor discovery is started
        before the ongoing discovery completes.

        Example usage::

            with ble.gap_connect(ble.ADDR_TYPE_PUBLIC, remote_mac) as conn:
                # Since characteristic discovery cannot be started while
                # another discovery is in process, we must consume the services
                # into a list before discovering their characteristics.
                services = list(conn.gattc_services())

                for service in services:
                    # Since descriptor discovery cannot be started while
                    # another discovery is in process, we must consume the
                    # characteristics into a list.
                    characteristics = list(conn.gattc_characteristics(service))

                    for characteristic in characteristics:
                        for descriptor in conn.gattc_descriptors(service, characteristic):
                            print(service, characteristic, descriptor)

        :param characteristic: A characteristic handle, or one of the characteristic
            tuples returned by ``gattc_characteristics()``.
        :return: An iterator of descriptor tuples, where each tuple contains
            a descriptor handle and a ``UUID`` value.
        :raises OSError: Descriptor discovery operation could not be started.
        """
        ...

    def gattc_read_characteristic(
        self,
        characteristic: Union[_CharacteristicHandle, _CharacteristicTuple],
        /,
    ) -> bytes:
        """
        Perform a GATT read operation against the specified characteristic
        of the connected BLE peripheral device.

        :param characteristic: A characteristic handle, or one of the characteristic
            tuples returned by ``gattc_characteristics()``.
        :return: A ``bytes`` object containing the characteristic value
            which was read.
        :raises OSError: The characteristic could not be read, or the
            connection was lost.
        """
        ...

    def gattc_read_descriptor(
        self,
        descriptor: Union[_DescriptorHandle, _DescriptorTuple],
        /,
    ) -> bytes:
        """
        Perform a GATT read operation against the specified descriptor
        of the connected BLE peripheral device.

        :param descriptor: A descriptor handle, or one of the descriptor
            tuples returned by ``gattc_descriptors()``.
        :return: A ``bytes`` object containing the descriptor value
            which was read.
        :raises OSError: The descriptor could not be read, or the
            connection was lost.
        """
        ...

    def gattc_write_characteristic(
        self,
        characteristic: Union[_CharacteristicHandle, _CharacteristicTuple],
        data: bytes,
        /,  # positional arguments only
    ) -> None:
        """
        Perform a GATT write operation against the specified characteristic
        of the connected BLE peripheral device.

        :param characteristic: A characteristic handle, or one of the characteristic
            tuples returned by ``gattc_characteristics()``.
        :param data: The value to be written to the remote device's characteristic.
            This parameter can be of type ``bytes`` or ``bytearray``.
        :raises OSError: The characteristic could not be written,
            or the connection was lost.
        """
        ...

    def gattc_write_descriptor(
        self,
        descriptor: Union[_DescriptorHandle, _DescriptorTuple],
        data: bytes,
        /,
    ) -> None:
        """
        Perform a GATT write operation against the specified descriptor
        of the connected BLE peripheral device.

        :param descriptor: A descriptor handle, or one of the descriptor
            tuples returned by ``gattc_descriptors()``.
        :param data: The value to be written to the remote device's descriptor.
            This parameter can be of type ``bytes`` or ``bytearray``.
        :raises OSError: The descriptor could not be written,
            or the connection was lost.
        """
        ...

    def gattc_configure(
        self,
        characteristic: Union[_CharacteristicHandle, _CharacteristicTuple],
        /,  # characteristic can only be specified as positional
        callback: Optional[Callable[[bytes, int], Any]] = None,
        *,
        notification: bool = False,
    ) -> None:
        """
        Enable or disable GATT notifications/indications for a given characteristic.

        This configures the remote GATT server to send notifications or indications
        on changes to the specified characteristic's value, and registers the
        given callback function to be called when a notification or indication
        is received.

        **Note:** Notifications are not acknowledged by the GATT client
        (in this case, the XBee) and do not guarantee delivery of data.

        :param characteristic: A characteristic handle, or one of the characteristic
            tuples returned by ``gattc_characteristics()``.
        :param callback: A function that is called whenever a notification or
            indication is received from the specified characteristic.
            This callback takes two parameters, a bytes object (data) and
            an integer (the offset of the data).
            If ``callback`` is None (the default), notifications/indications are
            disabled for the specified characteristic.
        :param notification: Optional parameter used to select whether to use
            notifications or indications. By default, indications are used.
            If ``notification`` is set to True, notifications are used.
        """
        ...

    def isconnected(self) -> bool:
        """
        Determines whether BLE is connected to a BLE peripheral device.

        :return: ``True`` if the BLE is connected to BLE peripheral device, ``False`` otherwise.
        """
        ...

    def secure(self, cb: Callable[[int], None], /) -> None:
        """Performs pairing/bonding on a connection.

        :param cb: A callback which will be called upon completion of
            the pairing operation. It will be passed the value zero if
            the pairing succeeded, otherwise it will be passed a BLE
            error as documented in the Bluetooth Core spec..

        See also the ``security`` argument to ``ble.config`` to guide
        the behavior of the pairing/bonding operation.

        The ``secure`` function is new in the following firmware versions:
            * XBee3 Cellular LTE-M/NB-IoT: version 11416
            * XBee3 Cellular LTE Cat 1: version x16

        """
        ...

    def __enter__(self) -> _gap_connect:
        """
        Enter the runtime context for using this GAP connection object as a context manager
        (using the ``with`` statement).

        This step has no effect, but the call to ``__exit__`` which occurs
        when this context is exited will have the same effect as ``close()``.

        You do not need to call ``__enter__`` directly, this happens automatically
        when the ``with`` statement is used.

        Example::

            with ble.gap_connect(ble.ADDR_TYPE_PUBLIC, addr) as conn:  # __enter__ is called
                for service in conn.gattc_services():
                    print(service)
            # ... context is exited. (The indented block under `with` is complete.)
            # __exit__ is called
            conn.gattc_services()  # raises OSError ENOTCONN

        :return: this GAP connection object
        """
        ...


def gap_connect(
        addr_type: int,
        address: bytes,
        /,  # addr_type and addr can only be specified as positional arguments
        timeout_ms: int = 5000,
        interval_us: int = 20000,
        window_us: int = 11250,
        onclose: Optional[Callable[[_gap_connect, int], Any]] = None
) -> _gap_connect:
    """Create a GAP connection to a BLE peripheral device.

    A GAP connection can be used to discover the services and characteristics
    supported by the connected peripheral device, and to interact with the
    device's characteristics by reading or writing values, or interacting with
    notifications and indications.

    **Close a connection:**

    To disconnect from a connected peripheral device, call the ``close()``
    method on the GAP connection object. Example::

        connection = ble.gap_connect(ble.ADDR_TYPE_PUBLIC, mac)
        # perform some activities with the connection
        connection.close()

    The GAP connection object will no longer be usable. Making a new connection
    requires calling ``gap_connect()`` again.

    **Using gap_connect as a context manager:**

    If a GAP connection object is left open (for example, if you connected
    to a device in a function but did not call ``close()``), resources in the
    XBee will continue to be used in servicing the connection until the
    connection object is deleted, either by soft-resetting the MicroPython REPL,
    or by garbage collection. (Note: Manually triggering garbage collection by
    using ``gc.collect()`` is typically not necessary in MicroPython.)

    Instead of needing to call ``close()`` directly, you may instead use the
    object returned by ``gap_connect`` as a context manager. By doing this,
    when the ``with`` block is exited, the GAP connection is automatically closed.
    This approach uses less code and is less error-prone, because ``close()`` will
    be called even if an exception is raised.

    For example, if you want to connect to a peripheral, discover and print its
    services, then close the connection, you can do this as follows::

        def show_services(addr_type, address):
            with ble.gap_connect(addr_type, address) as peripheral:
                for service in peripheral.gattc_services():
                    print(service)

        show_services(ble.ADDR_TYPE_PUBLIC, mac)

    Compare to this example which does not use a context manager. Note the use of
    a try/finally block, and the need to call ``close()`` explicitly::

        def show_services(addr_type, address):
            peripheral = ble.gap_connect(addr_type, address)
            try:
                for service in peripheral.gattc_services():
                    print(service)
            finally:
                # Make sure to call close() even if an exception is raised.
                peripheral.close()

        show_services(ble.ADDR_TYPE_PUBLIC, mac)

    **gap_scan and gap_connect:**

    Any ongoing GAP scan operation (see ``gap_scan()``) will block a GAP
    connection request until the scan completes. If the scan does not complete
    within ``timeout_ms``, the ``gap_connect`` call will raise OSError ETIMEDOUT.

    Starting a GAP scan operation while a GAP connection is open is allowed
    and does not affect the existing GAP connection.

    **timeout_ms parameter:**

    The ``timeout_ms`` parameter specifies the maximum length of time that the
    ``gap_connect`` call should be allowed to block waiting for the connection
    to be established. Note that per the Bluetooth Core Specification, if the
    remote device does not respond to a connection request within six (6)
    connection intervals, the connection shall be considered lost.
    In other words, the connection attempt can time out in less time than
    specified in ``timeout_ms``, depending on the connection interval used.
    See the ``interval_ms`` parameter for ``digi.ble.config()`` for more
    information on the connection interval setting.

    **Note:**

    The ``gap_connect`` function is new in the following firmware versions:
        * XBee3 Zigbee: version 100A
        * XBee3 802.15.4: version 200A
        * XBee3 Cellular LTE-M/NB-IoT: version 11415
        * XBee3 Cellular LTE Cat 1: version x15

    :param addr_type: The type of address contained in the ``address`` value.
        The possible values are defined as constants on the ``digi.ble`` module:
            * ``ADDR_TYPE_PUBLIC``
            * ``ADDR_TYPE_RANDOM``
            * ``ADDR_TYPE_PUBLIC_IDENTITY``
            * ``ADDR_TYPE_PUBLIC_RANDOM``
    :param address: The BLE MAC address to connect to. The address is a ``bytes`` object,
        and is 6 bytes (48 bits) long.
    :param timeout_ms: Specifies the maximum time to wait before giving up on
        a connection attempt.
    :param interval_us: Optionally configure the duty cycle of the GAP scan operation
        used to discover the remote device.
        The scanner will run for ``window_us`` microseconds every ``interval_us`` microseconds.
        ``window_us`` must be less than or equal to ``interval_us``.
        ``interval_us`` must be at least 2,500 microseconds (2.5 milliseconds)
        and no more than approximately 40.96 seconds (40,959,375 microseconds).
        The default interval is 20 milliseconds.
    :param window_us: Optionally configure the duty cycle of the GAP scan operation
        used to discover the remote device.
        The scanner will run for ``window_us`` microseconds every ``interval_us`` microseconds.
        ``window_us`` must be less than or equal to ``interval_us``.
        ``window_us`` must be at least 2,500 microseconds (2.5 milliseconds)
        and no more than approximately 40.96 seconds (40,959,375 microseconds).
        The default window is 11.25 milliseconds.
    :param onclose: When specified, indicates a function which will be called when
        the connection closes. The first argument to this callback is the
        connection returned by this function (gap_connect). The second
        argument is an error code. The error codes are taken from the
        Bluetooth Core specification Vol 2, Part D (Error Codes).

    :return: An object which encapsulates the GAP connection.
    :raises OSError ETIMEDOUT: The connection attempt timed out.
    :raises OSError ENOTCONN: The connection attempt failed for an unknown reason.

    """
    ...


class _xbee_connect():
    """
    Class used to encapsulate an encrypted connection to a remote XBee3 API Service.

    This class cannot be instantiated, it is returned from ``digi.ble.xbee_connect()``.

    See the documentation for ``digi.ble.xbee_connect()`` for examples of usage.
    """
    def send(self, data: bytes) -> None:
        """Send 'data' to the API Service of the peer.

        'data' should be valid API frames (including delimiter and
        checksum). Responses, if any, will be made through callbacks
        to the 'receive' argument used in the initial
        'digi.ble.xbee_connect()' call.

        """
        ...


def xbee_connect(conn: _gap_connect,
                 receive: Callable[[bytes], None],
                 password: str,
                 timeout: int) -> _xbee_connect:
    """Create an authenticated and encrypted connection to XBee3 API Service.

    This method will discover the API Service on the peer connection
    'conn' and authenticate using the provided password.

    Returns an object that allows encrypted communication using the
    state established by performing SRP.

    """
    ...


def delete_bondings() -> None:
    """Remove all stored bonding table entries."""
    ...


PasskeyCB = Callable[[int], None]
RequestCB = Callable[None, None]


def io_callbacks(display_cb: PasskeyCB,
                 confirm_cb: PasskeyCB,
                 request_cb: RequestCB
                 ) -> None:
    """Provide callbacks which define IO capabilities for pairing.

    :param display_cb: Callback to be used to present a passkey to the user.

    :param confirm_cb: Callback to be used when the user must
        confirm (Y/N) a passkey. The passkey will be provided as
        an argument. The passkey should be presented to the user
        and the users input should be fed back using
        ``ble.passkey_confirm``.

    :param request_cb: Callback to be indicate that the user must
        input a passkey value. The use should be prompted to enter
        a passkey and the passkey provided by the user should be
        fed back using ``ble.passkey_enter``.

    NOTE: The BLE standards recommend that the passkey be
    presented to the user as a six digit number padded with
    leading zeros.

    The ``io_callbacks`` function is new in the following firmware versions:
        * XBee3 Cellular LTE-M/NB-IoT: version 11416
        * XBee3 Cellular LTE Cat 1: version x16
    """
    ...


def passkey_confirm(confirmation: bool, /) -> None:
    """Allows user confirmation of BLE pairing passkey.

    :param confirmation: Provide True if the passkey provided by
        the ``confirm_cb`` is correct, False otherwise.

    The ``passkey_confirm`` function is new in the following firmware versions:
        * XBee3 Cellular LTE-M/NB-IoT: version 11416
        * XBee3 Cellular LTE Cat 1: version x16
    """
    ...


def passkey_enter(passkey: int, /) -> None:
    """Allows user entry of a passkey value.

    :param passkey: The numeric value of the passkey provided by
        the user.

    The ``passkey_enter`` function is new in the following firmware versions:
        * XBee3 Cellular LTE-M/NB-IoT: version 11416
        * XBee3 Cellular LTE Cat 1: version x16
    """
    ...
